import datetime
import importlib
import logging
import os
import sys
import time
from collections import Counter

import pytest

import sentry_sdk
import sentry_sdk.scope
from sentry_sdk import (
    Hub,
    add_breadcrumb,
    capture_event,
    capture_exception,
    capture_message,
    get_client,
    isolation_scope,
    last_event_id,
    new_scope,
    push_scope,
    start_transaction,
)
from sentry_sdk.client import Client
from sentry_sdk.integrations import (
    _AUTO_ENABLING_INTEGRATIONS,
    _DEFAULT_INTEGRATIONS,
    DidNotEnable,
    Integration,
    setup_integrations,
)
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.stdlib import StdlibIntegration
from sentry_sdk.scope import add_global_event_processor
from sentry_sdk.tracing_utils import has_tracing_enabled
from sentry_sdk.utils import datetime_from_isoformat, get_sdk_name, reraise


class NoOpIntegration(Integration):
    """
    A simple no-op integration for testing purposes.
    """

    identifier = "noop"

    @staticmethod
    def setup_once() -> None:
        pass

    def __eq__(self, __value: object) -> bool:
        """
        All instances of NoOpIntegration should be considered equal to each other.
        """
        return type(__value) == type(self)


def test_processors(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    def error_processor(event, exc_info):
        event["exception"]["values"][0]["value"] += " whatever"
        return event

    sentry_sdk.get_isolation_scope().add_error_processor(error_processor, ValueError)

    try:
        raise ValueError("aha!")
    except Exception:
        capture_exception()

    (event,) = events

    assert event["exception"]["values"][0]["value"] == "aha! whatever"


class ModuleImportErrorSimulator:
    def __init__(self, modules, error_cls=DidNotEnable):
        self.modules = modules
        self.error_cls = error_cls
        for sys_module in list(sys.modules.keys()):
            if any(sys_module.startswith(module) for module in modules):
                del sys.modules[sys_module]

    def find_spec(self, fullname, _path, _target=None):
        if fullname in self.modules:
            raise self.error_cls("Test import failure for %s" % fullname)

    def __enter__(self):
        # WARNING: We need to be first to avoid pytest messing with local imports
        sys.meta_path.insert(0, self)

    def __exit__(self, *_args):
        sys.meta_path.remove(self)


def test_auto_enabling_integrations_catches_import_error(sentry_init, caplog):
    caplog.set_level(logging.DEBUG)

    with ModuleImportErrorSimulator(
        [i.rsplit(".", 1)[0] for i in _AUTO_ENABLING_INTEGRATIONS]
    ):
        sentry_init(auto_enabling_integrations=True, debug=True)

    for import_string in _AUTO_ENABLING_INTEGRATIONS:
        assert any(
            record.message.startswith(
                "Did not import default integration {}:".format(import_string)
            )
            for record in caplog.records
        ), "Problem with checking auto enabling {}".format(import_string)


def test_generic_mechanism(sentry_init, capture_events):
    sentry_init()
    events = capture_events()

    try:
        raise ValueError("aha!")
    except Exception:
        capture_exception()

    (event,) = events
    assert event["exception"]["values"][0]["mechanism"]["type"] == "generic"
    assert event["exception"]["values"][0]["mechanism"]["handled"]


def test_option_before_send(sentry_init, capture_events):
    def before_send(event, hint):
        event["extra"] = {"before_send_called": True}
        return event

    def do_this():
        try:
            raise ValueError("aha!")
        except Exception:
            capture_exception()

    sentry_init(before_send=before_send)
    events = capture_events()

    do_this()

    (event,) = events
    assert event["extra"] == {"before_send_called": True}


def test_option_before_send_discard(sentry_init, capture_events):
    def before_send_discard(event, hint):
        return None

    def do_this():
        try:
            raise ValueError("aha!")
        except Exception:
            capture_exception()

    sentry_init(before_send=before_send_discard)
    events = capture_events()

    do_this()

    assert len(events) == 0


def test_option_before_send_transaction(sentry_init, capture_events):
    def before_send_transaction(event, hint):
        assert event["type"] == "transaction"
        event["extra"] = {"before_send_transaction_called": True}
        return event

    sentry_init(
        before_send_transaction=before_send_transaction,
        traces_sample_rate=1.0,
    )
    events = capture_events()
    transaction = start_transaction(name="foo")
    transaction.finish()

    (event,) = events
    assert event["transaction"] == "foo"
    assert event["extra"] == {"before_send_transaction_called": True}


def test_option_before_send_transaction_discard(sentry_init, capture_events):
    def before_send_transaction_discard(event, hint):
        return None

    sentry_init(
        before_send_transaction=before_send_transaction_discard,
        traces_sample_rate=1.0,
    )
    events = capture_events()
    transaction = start_transaction(name="foo")
    transaction.finish()

    assert len(events) == 0


def test_option_before_breadcrumb(sentry_init, capture_events, monkeypatch):
    drop_events = False
    drop_breadcrumbs = False
    reports = []

    def record_lost_event(reason, data_category=None, item=None):
        reports.append((reason, data_category))

    def before_send(event, hint):
        assert isinstance(hint["exc_info"][1], ValueError)
        if not drop_events:
            event["extra"] = {"foo": "bar"}
            return event

    def before_breadcrumb(crumb, hint):
        assert hint == {"foo": 42}
        if not drop_breadcrumbs:
            crumb["data"] = {"foo": "bar"}
            return crumb

    sentry_init(before_send=before_send, before_breadcrumb=before_breadcrumb)
    events = capture_events()

    monkeypatch.setattr(
        sentry_sdk.get_client().transport, "record_lost_event", record_lost_event
    )

    def do_this():
        add_breadcrumb(message="Hello", hint={"foo": 42})
        try:
            raise ValueError("aha!")
        except Exception:
            capture_exception()

    do_this()
    drop_breadcrumbs = True
    do_this()
    assert not reports
    drop_events = True
    do_this()
    assert reports == [("before_send", "error")]

    normal, no_crumbs = events

    assert normal["exception"]["values"][0]["type"] == "ValueError"
    (crumb,) = normal["breadcrumbs"]["values"]
    assert "timestamp" in crumb
    assert crumb["message"] == "Hello"
    assert crumb["data"] == {"foo": "bar"}
    assert crumb["type"] == "default"


@pytest.mark.parametrize(
    "enable_tracing, traces_sample_rate, tracing_enabled, updated_traces_sample_rate",
    [
        (None, None, False, None),
        (False, 0.0, False, 0.0),
        (False, 1.0, False, 1.0),
        (None, 1.0, True, 1.0),
        (True, 1.0, True, 1.0),
        (None, 0.0, True, 0.0),  # We use this as - it's configured but turned off
        (True, 0.0, True, 0.0),  # We use this as - it's configured but turned off
        (True, None, True, 1.0),
    ],
)
def test_option_enable_tracing(
    sentry_init,
    enable_tracing,
    traces_sample_rate,
    tracing_enabled,
    updated_traces_sample_rate,
):
    sentry_init(enable_tracing=enable_tracing, traces_sample_rate=traces_sample_rate)
    options = sentry_sdk.get_client().options
    assert has_tracing_enabled(options) is tracing_enabled
    assert options["traces_sample_rate"] == updated_traces_sample_rate


def test_breadcrumb_arguments(sentry_init, capture_events):
    assert_hint = {"bar": 42}

    def before_breadcrumb(crumb, hint):
        assert crumb["foo"] == 42
        assert hint == assert_hint

    sentry_init(before_breadcrumb=before_breadcrumb)

    add_breadcrumb(foo=42, hint=dict(bar=42))
    add_breadcrumb(dict(foo=42), dict(bar=42))
    add_breadcrumb(dict(foo=42), hint=dict(bar=42))
    add_breadcrumb(crumb=dict(foo=42), hint=dict(bar=42))

    assert_hint.clear()
    add_breadcrumb(foo=42)
    add_breadcrumb(crumb=dict(foo=42))


def test_push_scope(sentry_init, capture_events, suppress_deprecation_warnings):
    sentry_init()
    events = capture_events()

    with push_scope() as scope:
        scope.level = "warning"
        try:
            1 / 0
        except Exception as e:
            capture_exception(e)

    (event,) = events

    assert event["level"] == "warning"
    assert "exception" in event


def test_push_scope_null_client(
    sentry_init, capture_events, suppress_deprecation_warnings
):
    """
    This test can be removed when we remove push_scope and the Hub from the SDK.
    """
    sentry_init()
    events = capture_events()

    Hub.current.bind_client(None)

    with push_scope() as scope:
        scope.level = "warning"
        try:
            1 / 0
        except Exception as e:
            capture_exception(e)

    assert len(events) == 0


@pytest.mark.skip(
    reason="This test is not valid anymore, because push_scope just returns the isolation scope. This test should be removed once the Hub is removed"
)
@pytest.mark.parametrize("null_client", (True, False))
def test_push_scope_callback(sentry_init, null_client, capture_events):
    """
    This test can be removed when we remove push_scope and the Hub from the SDK.
    """
    sentry_init()

    if null_client:
        Hub.current.bind_client(None)

    outer_scope = Hub.current.scope

    calls = []

    @push_scope
    def _(scope):
        assert scope is Hub.current.scope
        assert scope is not outer_scope
        calls.append(1)

    # push_scope always needs to execute the callback regardless of
    # client state, because that actually runs usercode in it, not
    # just scope config code
    assert calls == [1]

    # Assert scope gets popped correctly
    assert Hub.current.scope is outer_scope


def test_breadcrumbs(sentry_init, capture_events):
    sentry_init(max_breadcrumbs=10)
    events = capture_events()

    for i in range(20):
        add_breadcrumb(
            category="auth", message="Authenticated user %s" % i, level="info"
        )

    capture_exception(ValueError())
    (event,) = events

    assert len(event["breadcrumbs"]["values"]) == 10
    assert "user 10" in event["breadcrumbs"]["values"][0]["message"]
    assert "user 19" in event["breadcrumbs"]["values"][-1]["message"]

    del events[:]

    for i in range(2):
        add_breadcrumb(
            category="auth", message="Authenticated user %s" % i, level="info"
        )

    sentry_sdk.get_isolation_scope().clear()

    capture_exception(ValueError())
    (event,) = events
    assert len(event["breadcrumbs"]["values"]) == 0


def test_breadcrumb_ordering(sentry_init, capture_events):
    sentry_init()
    events = capture_events()
    now = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)

    timestamps = [
        now - datetime.timedelta(days=10),
        now - datetime.timedelta(days=8),
        now - datetime.timedelta(days=12),
    ]

    for timestamp in timestamps:
        add_breadcrumb(
            message="Authenticated at %s" % timestamp,
            category="auth",
            level="info",
            timestamp=timestamp,
        )

    capture_exception(ValueError())
    (event,) = events

    assert len(event["breadcrumbs"]["values"]) == len(timestamps)
    timestamps_from_event = [
        datetime_from_isoformat(x["timestamp"]) for x in event["breadcrumbs"]["values"]
    ]
    assert timestamps_from_event == sorted(timestamps)


def test_breadcrumb_ordering_different_types(sentry_init, capture_events):
    sentry_init()
    events = capture_events()
    now = datetime.datetime.now(datetime.timezone.utc)

    timestamps = [
        now - datetime.timedelta(days=10),
        now - datetime.timedelta(days=8),
        now.replace(microsecond=0) - datetime.timedelta(days=12),
        now - datetime.timedelta(days=9),
        now - datetime.timedelta(days=13),
        now.replace(microsecond=0) - datetime.timedelta(days=11),
    ]

    breadcrumb_timestamps = [
        timestamps[0],
        timestamps[1].isoformat(),
        datetime.datetime.strftime(timestamps[2], "%Y-%m-%dT%H:%M:%S") + "Z",
        datetime.datetime.strftime(timestamps[3], "%Y-%m-%dT%H:%M:%S.%f") + "+00:00",
        datetime.datetime.strftime(timestamps[4], "%Y-%m-%dT%H:%M:%S.%f") + "+0000",
        datetime.datetime.strftime(timestamps[5], "%Y-%m-%dT%H:%M:%S.%f") + "-0000",
    ]

    for i, timestamp in enumerate(timestamps):
        add_breadcrumb(
            message="Authenticated at %s" % timestamp,
            category="auth",
            level="info",
            timestamp=breadcrumb_timestamps[i],
        )

    capture_exception(ValueError())
    (event,) = events

    assert len(event["breadcrumbs"]["values"]) == len(timestamps)
    timestamps_from_event = [
        datetime_from_isoformat(x["timestamp"]) for x in event["breadcrumbs"]["values"]
    ]
    assert timestamps_from_event == sorted(timestamps)


def test_attachments(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    this_file = os.path.abspath(__file__.rstrip("c"))

    scope = sentry_sdk.get_isolation_scope()
    scope.add_attachment(bytes=b"Hello World!", filename="message.txt")
    scope.add_attachment(path=this_file)

    capture_exception(ValueError())

    (envelope,) = envelopes

    assert len(envelope.items) == 3
    assert envelope.get_event()["exception"] is not None

    attachments = [x for x in envelope.items if x.type == "attachment"]
    (message, pyfile) = attachments

    assert message.headers["filename"] == "message.txt"
    assert message.headers["type"] == "attachment"
    assert message.headers["content_type"] == "text/plain"
    assert message.payload.bytes == message.payload.get_bytes() == b"Hello World!"

    assert pyfile.headers["filename"] == os.path.basename(this_file)
    assert pyfile.headers["type"] == "attachment"
    assert pyfile.headers["content_type"].startswith("text/")
    assert pyfile.payload.bytes is None
    with open(this_file, "rb") as f:
        assert pyfile.payload.get_bytes() == f.read()


@pytest.mark.tests_internal_exceptions
def test_attachments_graceful_failure(
    sentry_init, capture_envelopes, internal_exceptions
):
    sentry_init()
    envelopes = capture_envelopes()

    sentry_sdk.get_isolation_scope().add_attachment(path="non_existent")
    capture_exception(ValueError())

    (envelope,) = envelopes
    assert len(envelope.items) == 2
    assert envelope.items[1].payload.get_bytes() == b""


def test_attachments_exceptions(sentry_init):
    sentry_init()

    scope = sentry_sdk.get_isolation_scope()

    # bytes and path are None
    with pytest.raises(TypeError) as e:
        scope.add_attachment()

    assert str(e.value) == "path or raw bytes required for attachment"

    # filename is None
    with pytest.raises(TypeError) as e:
        scope.add_attachment(bytes=b"Hello World!")

    assert str(e.value) == "filename is required for attachment"


def test_attachments_content_type_is_none(sentry_init, capture_envelopes):
    sentry_init()
    envelopes = capture_envelopes()

    scope = sentry_sdk.get_isolation_scope()

    scope.add_attachment(
        bytes=b"Hello World!", filename="message.txt", content_type="foo/bar"
    )
    capture_exception(ValueError())

    (envelope,) = envelopes
    attachments = [x for x in envelope.items if x.type == "attachment"]
    (message,) = attachments

    assert message.headers["filename"] == "message.txt"
    assert message.headers["content_type"] == "foo/bar"


def test_attachments_repr(sentry_init):
    sentry_init()

    scope = sentry_sdk.get_isolation_scope()

    scope.add_attachment(bytes=b"Hello World!", filename="message.txt")

    assert repr(scope._attachments[0]) == "<Attachment 'message.txt'>"


def test_attachments_bytes_callable_payload(sentry_init):
    sentry_init()

    scope = sentry_sdk.get_isolation_scope()

    scope.add_attachment(bytes=bytes, filename="message.txt")

    attachment = scope._attachments[0]
    item = attachment.to_envelope_item()

    assert item.payload.bytes == b""


def test_integration_scoping(sentry_init, capture_events):
    logger = logging.getLogger("test_basics")

    # This client uses the logging integration
    logging_integration = LoggingIntegration(event_level=logging.WARNING)
    sentry_init(default_integrations=False, integrations=[logging_integration])
    events = capture_events()
    logger.warning("This is a warning")
    assert len(events) == 1

    # This client does not
    sentry_init(default_integrations=False)
    events = capture_events()
    logger.warning("This is not a warning")
    assert not events


default_integrations = [
    getattr(
        importlib.import_module(integration.rsplit(".", 1)[0]),
        integration.rsplit(".", 1)[1],
    )
    for integration in _DEFAULT_INTEGRATIONS
]


@pytest.mark.forked
@pytest.mark.parametrize(
    "provided_integrations,default_integrations,disabled_integrations,expected_integrations",
    [
        ([], False, None, set()),
        ([], False, [], set()),
        ([LoggingIntegration()], False, None, {LoggingIntegration}),
        ([], True, None, set(default_integrations)),
        (
            [],
            True,
            [LoggingIntegration(), StdlibIntegration],
            set(default_integrations) - {LoggingIntegration, StdlibIntegration},
        ),
    ],
)
def test_integrations(
    sentry_init,
    provided_integrations,
    default_integrations,
    disabled_integrations,
    expected_integrations,
    reset_integrations,
):
    sentry_init(
        integrations=provided_integrations,
        default_integrations=default_integrations,
        disabled_integrations=disabled_integrations,
        auto_enabling_integrations=False,
        debug=True,
    )
    assert {
        type(integration) for integration in get_client().integrations.values()
    } == expected_integrations


@pytest.mark.skip(
    reason="This test is not valid anymore, because with the new Scopes calling bind_client on the Hub sets the client on the global scope. This test should be removed once the Hub is removed"
)
def test_client_initialized_within_scope(sentry_init, caplog):
    """
    This test can be removed when we remove push_scope and the Hub from the SDK.
    """
    caplog.set_level(logging.WARNING)

    sentry_init()

    with push_scope():
        Hub.current.bind_client(Client())

    (record,) = (x for x in caplog.records if x.levelname == "WARNING")

    assert record.msg.startswith("init() called inside of pushed scope.")


@pytest.mark.skip(
    reason="This test is not valid anymore, because with the new Scopes the push_scope just returns the isolation scope. This test should be removed once the Hub is removed"
)
def test_scope_leaks_cleaned_up(sentry_init, caplog):
    """
    This test can be removed when we remove push_scope and the Hub from the SDK.
    """
    caplog.set_level(logging.WARNING)

    sentry_init()

    old_stack = list(Hub.current._stack)

    with push_scope():
        push_scope()

    assert Hub.current._stack == old_stack

    (record,) = (x for x in caplog.records if x.levelname == "WARNING")

    assert record.message.startswith("Leaked 1 scopes:")


@pytest.mark.skip(
    reason="This test is not valid anymore, because with the new Scopes there is not pushing and popping of scopes. This test should be removed once the Hub is removed"
)
def test_scope_popped_too_soon(sentry_init, caplog):
    """
    This test can be removed when we remove push_scope and the Hub from the SDK.
    """
    caplog.set_level(logging.ERROR)

    sentry_init()

    old_stack = list(Hub.current._stack)

    with push_scope():
        Hub.current.pop_scope_unsafe()

    assert Hub.current._stack == old_stack

    (record,) = (x for x in caplog.records if x.levelname == "ERROR")

    assert record.message == ("Scope popped too soon. Popped 1 scopes too many.")


def test_scope_event_processor_order(sentry_init, capture_events):
    def before_send(event, hint):
        event["message"] += "baz"
        return event

    sentry_init(debug=True, before_send=before_send)
    events = capture_events()

    with new_scope() as scope:

        @scope.add_event_processor
        def foo(event, hint):
            event["message"] += "foo"
            return event

        with new_scope() as scope:

            @scope.add_event_processor
            def bar(event, hint):
                event["message"] += "bar"
                return event

            capture_message("hi")

    (event,) = events

    assert event["message"] == "hifoobarbaz"


def test_capture_event_with_scope_kwargs(sentry_init, capture_events):
    sentry_init()
    events = capture_events()
    capture_event({}, level="info", extras={"foo": "bar"})
    (event,) = events
    assert event["level"] == "info"
    assert event["extra"]["foo"] == "bar"


def test_dedupe_event_processor_drop_records_client_report(
    sentry_init, capture_events, capture_record_lost_event_calls
):
    """
    DedupeIntegration internally has an event_processor that filters duplicate exceptions.
    We want a duplicate exception to be captured only once and the drop being recorded as
    a client report.
    """
    sentry_init()
    events = capture_events()
    record_lost_event_calls = capture_record_lost_event_calls()

    try:
        raise ValueError("aha!")
    except Exception:
        try:
            capture_exception()
            reraise(*sys.exc_info())
        except Exception:
            capture_exception()

    (event,) = events
    (lost_event_call,) = record_lost_event_calls

    assert event["level"] == "error"
    assert "exception" in event
    assert lost_event_call == ("event_processor", "error", None, 1)


def test_dedupe_doesnt_take_into_account_dropped_exception(sentry_init, capture_events):
    # Two exceptions happen one after another. The first one is dropped in the
    # user's before_send. The second one isn't.
    # Originally, DedupeIntegration would drop the second exception. This test
    # is making sure that that is no longer the case -- i.e., DedupeIntegration
    # doesn't consider exceptions dropped in before_send.
    count = 0

    def before_send(event, hint):
        nonlocal count
        count += 1
        if count == 1:
            return None
        return event

    sentry_init(before_send=before_send)
    events = capture_events()

    exc = ValueError("aha!")
    for _ in range(2):
        # The first ValueError will be dropped by before_send. The second
        # ValueError will be accepted by before_send, and should be sent to
        # Sentry.
        try:
            raise exc
        except Exception:
            capture_exception()

    assert len(events) == 1


def test_event_processor_drop_records_client_report(
    sentry_init, capture_events, capture_record_lost_event_calls
):
    sentry_init(traces_sample_rate=1.0)
    events = capture_events()
    record_lost_event_calls = capture_record_lost_event_calls()

    # Ensure full idempotency by restoring the original global event processors list object, not just a copy.
    old_processors = sentry_sdk.scope.global_event_processors

    try:
        sentry_sdk.scope.global_event_processors = (
            sentry_sdk.scope.global_event_processors.copy()
        )

        @add_global_event_processor
        def foo(event, hint):
            return None

        capture_message("dropped")

        with start_transaction(name="dropped"):
            pass

        assert len(events) == 0

        # Using Counter because order of record_lost_event calls does not matter
        assert Counter(record_lost_event_calls) == Counter(
            [
                ("event_processor", "error", None, 1),
                ("event_processor", "transaction", None, 1),
                ("event_processor", "span", None, 1),
            ]
        )

    finally:
        sentry_sdk.scope.global_event_processors = old_processors


@pytest.mark.parametrize(
    "installed_integrations, expected_name",
    [
        # integrations with own name
        (["django"], "sentry.python.django"),
        (["flask"], "sentry.python.flask"),
        (["fastapi"], "sentry.python.fastapi"),
        (["bottle"], "sentry.python.bottle"),
        (["falcon"], "sentry.python.falcon"),
        (["quart"], "sentry.python.quart"),
        (["sanic"], "sentry.python.sanic"),
        (["starlette"], "sentry.python.starlette"),
        (["starlite"], "sentry.python.starlite"),
        (["litestar"], "sentry.python.litestar"),
        (["chalice"], "sentry.python.chalice"),
        (["serverless"], "sentry.python.serverless"),
        (["pyramid"], "sentry.python.pyramid"),
        (["tornado"], "sentry.python.tornado"),
        (["aiohttp"], "sentry.python.aiohttp"),
        (["aws_lambda"], "sentry.python.aws_lambda"),
        (["gcp"], "sentry.python.gcp"),
        (["beam"], "sentry.python.beam"),
        (["asgi"], "sentry.python.asgi"),
        (["wsgi"], "sentry.python.wsgi"),
        # integrations without name
        (["argv"], "sentry.python"),
        (["atexit"], "sentry.python"),
        (["boto3"], "sentry.python"),
        (["celery"], "sentry.python"),
        (["dedupe"], "sentry.python"),
        (["excepthook"], "sentry.python"),
        (["unraisablehook"], "sentry.python"),
        (["executing"], "sentry.python"),
        (["modules"], "sentry.python"),
        (["pure_eval"], "sentry.python"),
        (["redis"], "sentry.python"),
        (["rq"], "sentry.python"),
        (["sqlalchemy"], "sentry.python"),
        (["stdlib"], "sentry.python"),
        (["threading"], "sentry.python"),
        (["trytond"], "sentry.python"),
        (["logging"], "sentry.python"),
        (["gnu_backtrace"], "sentry.python"),
        (["httpx"], "sentry.python"),
        # precedence of frameworks
        (["flask", "django", "celery"], "sentry.python.django"),
        (["fastapi", "flask", "redis"], "sentry.python.flask"),
        (["bottle", "fastapi", "httpx"], "sentry.python.fastapi"),
        (["falcon", "bottle", "logging"], "sentry.python.bottle"),
        (["quart", "falcon", "gnu_backtrace"], "sentry.python.falcon"),
        (["sanic", "quart", "sqlalchemy"], "sentry.python.quart"),
        (["starlette", "sanic", "rq"], "sentry.python.sanic"),
        (["chalice", "starlette", "modules"], "sentry.python.starlette"),
        (["chalice", "starlite", "modules"], "sentry.python.starlite"),
        (["chalice", "litestar", "modules"], "sentry.python.litestar"),
        (["serverless", "chalice", "pure_eval"], "sentry.python.chalice"),
        (["pyramid", "serverless", "modules"], "sentry.python.serverless"),
        (["tornado", "pyramid", "executing"], "sentry.python.pyramid"),
        (["aiohttp", "tornado", "dedupe"], "sentry.python.tornado"),
        (["aws_lambda", "aiohttp", "boto3"], "sentry.python.aiohttp"),
        (["gcp", "aws_lambda", "atexit"], "sentry.python.aws_lambda"),
        (["beam", "gcp", "argv"], "sentry.python.gcp"),
        (["asgi", "beam", "stdtlib"], "sentry.python.beam"),
        (["wsgi", "asgi", "boto3"], "sentry.python.asgi"),
        (["wsgi", "celery", "redis"], "sentry.python.wsgi"),
    ],
)
def test_get_sdk_name(installed_integrations, expected_name):
    assert get_sdk_name(installed_integrations) == expected_name


def _hello_world(word):
    return "Hello, {}".format(word)


def test_functions_to_trace(sentry_init, capture_events):
    functions_to_trace = [
        {"qualified_name": "tests.test_basics._hello_world"},
        {"qualified_name": "time.sleep"},
    ]

    sentry_init(
        traces_sample_rate=1.0,
        functions_to_trace=functions_to_trace,
    )

    events = capture_events()

    with start_transaction(name="something"):
        time.sleep(0)

        for word in ["World", "You"]:
            _hello_world(word)

    assert len(events) == 1

    (event,) = events

    assert len(event["spans"]) == 3
    assert event["spans"][0]["description"] == "time.sleep"
    assert event["spans"][1]["description"] == "tests.test_basics._hello_world"
    assert event["spans"][2]["description"] == "tests.test_basics._hello_world"


class WorldGreeter:
    def __init__(self, word):
        self.word = word

    def greet(self, new_word=None):
        return "Hello, {}".format(new_word if new_word else self.word)


def test_functions_to_trace_with_class(sentry_init, capture_events):
    functions_to_trace = [
        {"qualified_name": "tests.test_basics.WorldGreeter.greet"},
    ]

    sentry_init(
        traces_sample_rate=1.0,
        functions_to_trace=functions_to_trace,
    )

    events = capture_events()

    with start_transaction(name="something"):
        wg = WorldGreeter("World")
        wg.greet()
        wg.greet("You")

    assert len(events) == 1

    (event,) = events

    assert len(event["spans"]) == 2
    assert event["spans"][0]["description"] == "tests.test_basics.WorldGreeter.greet"
    assert event["spans"][1]["description"] == "tests.test_basics.WorldGreeter.greet"


def test_multiple_setup_integrations_calls():
    first_call_return = setup_integrations([NoOpIntegration()], with_defaults=False)
    assert first_call_return == {NoOpIntegration.identifier: NoOpIntegration()}

    second_call_return = setup_integrations([NoOpIntegration()], with_defaults=False)
    assert second_call_return == {NoOpIntegration.identifier: NoOpIntegration()}


class TracingTestClass:
    @staticmethod
    def static(arg):
        return arg

    @classmethod
    def class_(cls, arg):
        return cls, arg


# We need to fork here because the test modifies tests.test_basics.TracingTestClass
@pytest.mark.forked
def test_staticmethod_class_tracing(sentry_init, capture_events):
    sentry_init(
        debug=True,
        traces_sample_rate=1.0,
        functions_to_trace=[
            {"qualified_name": "tests.test_basics.TracingTestClass.static"}
        ],
    )

    events = capture_events()

    with sentry_sdk.start_transaction(name="test"):
        assert TracingTestClass.static(1) == 1

    (event,) = events
    assert event["type"] == "transaction"
    assert event["transaction"] == "test"

    (span,) = event["spans"]
    assert span["description"] == "tests.test_basics.TracingTestClass.static"


# We need to fork here because the test modifies tests.test_basics.TracingTestClass
@pytest.mark.forked
def test_staticmethod_instance_tracing(sentry_init, capture_events):
    sentry_init(
        debug=True,
        traces_sample_rate=1.0,
        functions_to_trace=[
            {"qualified_name": "tests.test_basics.TracingTestClass.static"}
        ],
    )

    events = capture_events()

    with sentry_sdk.start_transaction(name="test"):
        assert TracingTestClass().static(1) == 1

    (event,) = events
    assert event["type"] == "transaction"
    assert event["transaction"] == "test"

    (span,) = event["spans"]
    assert span["description"] == "tests.test_basics.TracingTestClass.static"


# We need to fork here because the test modifies tests.test_basics.TracingTestClass
@pytest.mark.forked
def test_classmethod_class_tracing(sentry_init, capture_events):
    sentry_init(
        debug=True,
        traces_sample_rate=1.0,
        functions_to_trace=[
            {"qualified_name": "tests.test_basics.TracingTestClass.class_"}
        ],
    )

    events = capture_events()

    with sentry_sdk.start_transaction(name="test"):
        assert TracingTestClass.class_(1) == (TracingTestClass, 1)

    (event,) = events
    assert event["type"] == "transaction"
    assert event["transaction"] == "test"

    (span,) = event["spans"]
    assert span["description"] == "tests.test_basics.TracingTestClass.class_"


# We need to fork here because the test modifies tests.test_basics.TracingTestClass
@pytest.mark.forked
def test_classmethod_instance_tracing(sentry_init, capture_events):
    sentry_init(
        debug=True,
        traces_sample_rate=1.0,
        functions_to_trace=[
            {"qualified_name": "tests.test_basics.TracingTestClass.class_"}
        ],
    )

    events = capture_events()

    with sentry_sdk.start_transaction(name="test"):
        assert TracingTestClass().class_(1) == (TracingTestClass, 1)

    (event,) = events
    assert event["type"] == "transaction"
    assert event["transaction"] == "test"

    (span,) = event["spans"]
    assert span["description"] == "tests.test_basics.TracingTestClass.class_"


def test_functions_to_trace_no_dot_does_not_crash(sentry_init):
    # A qualified_name with no dot should not raise ValueError during sentry_sdk.init()
    sentry_init(
        traces_sample_rate=1.0,
        functions_to_trace=[{"qualified_name": "my_function"}],
    )


def test_last_event_id(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    assert last_event_id() is None

    capture_exception(Exception("test"))

    assert last_event_id() is not None


def test_last_event_id_transaction(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    assert last_event_id() is None

    with start_transaction(name="test"):
        pass

    assert last_event_id() is None, "Transaction should not set last_event_id"


def test_last_event_id_scope(sentry_init):
    sentry_init(traces_sample_rate=1.0)

    # Should not crash
    with isolation_scope() as scope:
        assert scope.last_event_id() is None


def test_hub_constructor_deprecation_warning():
    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning):
        Hub()


def test_hub_current_deprecation_warning():
    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning) as warning_records:
        Hub.current

    # Make sure we only issue one deprecation warning
    assert len(warning_records) == 1


def test_hub_main_deprecation_warnings():
    with pytest.warns(sentry_sdk.hub.SentryHubDeprecationWarning):
        Hub.main


@pytest.mark.skipif(sys.version_info < (3, 11), reason="add_note() not supported")
def test_notes(sentry_init, capture_events):
    sentry_init()
    events = capture_events()
    try:
        e = ValueError("aha!")
        e.add_note("Test 123")
        e.add_note("another note")
        raise e
    except Exception:
        capture_exception()

    (event,) = events

    assert event["exception"]["values"][0]["value"] == "aha!\nTest 123\nanother note"


@pytest.mark.skipif(sys.version_info < (3, 11), reason="add_note() not supported")
def test_notes_safe_str(sentry_init, capture_events):
    class Note2:
        def __repr__(self):
            raise TypeError

        def __str__(self):
            raise TypeError

    sentry_init()
    events = capture_events()
    try:
        e = ValueError("aha!")
        e.add_note("note 1")
        e.__notes__.append(Note2())  # type: ignore
        e.add_note("note 3")
        e.__notes__.append(2)  # type: ignore
        raise e
    except Exception:
        capture_exception()

    (event,) = events

    assert event["exception"]["values"][0]["value"] == "aha!\nnote 1\nnote 3"


@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="this test appears to cause a segfault on Python < 3.11",
)
def test_stacktrace_big_recursion(sentry_init, capture_events):
    """
    Ensure that if the recursion limit is increased, the full stacktrace is not captured,
    as it would take too long to process the entire stack trace.
    Also, ensure that the capturing does not take too long.
    """
    sentry_init()
    events = capture_events()

    def recurse():
        recurse()

    old_recursion_limit = sys.getrecursionlimit()

    try:
        sys.setrecursionlimit(100_000)
        recurse()
    except RecursionError as e:
        capture_start_time = time.perf_counter_ns()
        sentry_sdk.capture_exception(e)
        capture_end_time = time.perf_counter_ns()
    finally:
        sys.setrecursionlimit(old_recursion_limit)

    (event,) = events

    assert event["exception"]["values"][0]["stacktrace"] is None
    assert event["_meta"]["exception"] == {
        "values": {"0": {"stacktrace": {"": {"rem": [["!config", "x"]]}}}}
    }

    # On my machine, it takes about 100-200ms to capture the exception,
    # so this limit should be generous enough.
    assert capture_end_time - capture_start_time < 10**9 * 2, (
        "stacktrace capture took too long, check that frame limit is set correctly"
    )



import pytest

import sentry_sdk
from sentry_sdk.consts import OP
from sentry_sdk.integrations.django import DjangoIntegration

try:
    from django.tasks import task

    HAS_DJANGO_TASKS = True
except ImportError:
    HAS_DJANGO_TASKS = False


@pytest.fixture
def immediate_backend(settings):
    """Configure Django to use the immediate task backend for synchronous testing."""
    settings.TASKS = {
        "default": {"BACKEND": "django.tasks.backends.immediate.ImmediateBackend"}
    }


if HAS_DJANGO_TASKS:

    @task
    def simple_task():
        return "result"

    @task
    def add_numbers(a, b):
        return a + b

    @task
    def greet(name, greeting="Hello"):
        return f"{greeting}, {name}!"

    @task
    def failing_task():
        raise ValueError("Task failed!")

    @task
    def task_one():
        return 1

    @task
    def task_two():
        return 2


@pytest.mark.skipif(
    not HAS_DJANGO_TASKS,
    reason="Django tasks are only available in Django 6.0+",
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_task_span_is_created(
    sentry_init,
    capture_events,
    capture_items,
    immediate_backend,
    span_streaming,
):
    """Test that the queue.submit.django span is created when a task is enqueued."""
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            simple_task.enqueue()

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        queue_submit_spans = [
            span
            for span in spans
            if span["attributes"].get("sentry.op") == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["name"]
            == "tests.integrations.django.test_tasks.simple_task"
        )
        assert (
            queue_submit_spans[0]["attributes"]["sentry.origin"] == "auto.http.django"
        )
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction"):
            simple_task.enqueue()

        (event,) = events
        assert event["type"] == "transaction"

        queue_submit_spans = [
            span for span in event["spans"] if span["op"] == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["description"]
            == "tests.integrations.django.test_tasks.simple_task"
        )
        assert queue_submit_spans[0]["origin"] == "auto.http.django"


@pytest.mark.skipif(
    not HAS_DJANGO_TASKS,
    reason="Django tasks are only available in Django 6.0+",
)
def test_task_enqueue_returns_result(sentry_init, immediate_backend):
    """Test that the task enqueuing behavior is unchanged from the user perspective."""
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
    )

    result = add_numbers.enqueue(3, 5)

    assert result is not None
    assert result.return_value == 8


@pytest.mark.skipif(
    not HAS_DJANGO_TASKS,
    reason="Django tasks are only available in Django 6.0+",
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_task_enqueue_with_kwargs(
    sentry_init,
    immediate_backend,
    capture_events,
    capture_items,
    span_streaming,
):
    """Test that task enqueuing works correctly with keyword arguments."""
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            result = greet.enqueue(name="World", greeting="Hi")

        assert result.return_value == "Hi, World!"

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        queue_submit_spans = [
            span
            for span in spans
            if span["attributes"].get("sentry.op") == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["name"]
            == "tests.integrations.django.test_tasks.greet"
        )
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction"):
            result = greet.enqueue(name="World", greeting="Hi")

        assert result.return_value == "Hi, World!"

        (event,) = events
        queue_submit_spans = [
            span for span in event["spans"] if span["op"] == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["description"]
            == "tests.integrations.django.test_tasks.greet"
        )


@pytest.mark.skipif(
    not HAS_DJANGO_TASKS,
    reason="Django tasks are only available in Django 6.0+",
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_task_error_reporting(
    sentry_init,
    immediate_backend,
    capture_events,
    capture_items,
    span_streaming,
):
    """Test that errors in tasks are correctly reported and don't break the span."""
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            result = failing_task.enqueue()

        with pytest.raises(ValueError, match="Task failed"):
            _ = result.return_value

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        queue_submit_spans = [
            span
            for span in spans
            if span["attributes"].get("sentry.op") == OP.QUEUE_SUBMIT_DJANGO
        ]

        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["name"]
            == "tests.integrations.django.test_tasks.failing_task"
        )
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction"):
            result = failing_task.enqueue()

        with pytest.raises(ValueError, match="Task failed"):
            _ = result.return_value

        assert len(events) == 2
        transaction_event = events[-1]
        assert transaction_event["type"] == "transaction"

        queue_submit_spans = [
            span
            for span in transaction_event["spans"]
            if span["op"] == OP.QUEUE_SUBMIT_DJANGO
        ]

        assert len(queue_submit_spans) == 1
        assert (
            queue_submit_spans[0]["description"]
            == "tests.integrations.django.test_tasks.failing_task"
        )


@pytest.mark.skipif(
    not HAS_DJANGO_TASKS,
    reason="Django tasks are only available in Django 6.0+",
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_multiple_task_enqueues_create_multiple_spans(
    sentry_init,
    capture_events,
    capture_items,
    immediate_backend,
    span_streaming,
):
    """Test that enqueueing multiple tasks creates multiple spans."""
    sentry_init(
        integrations=[DjangoIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            task_one.enqueue()
            task_two.enqueue()
            task_one.enqueue()

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        queue_submit_spans = [
            span
            for span in spans
            if span["attributes"].get("sentry.op") == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 3

        span_names = [span["name"] for span in queue_submit_spans]
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(name="test_transaction"):
            task_one.enqueue()
            task_two.enqueue()
            task_one.enqueue()

        (event,) = events
        queue_submit_spans = [
            span for span in event["spans"] if span["op"] == OP.QUEUE_SUBMIT_DJANGO
        ]
        assert len(queue_submit_spans) == 3

        span_names = [span["description"] for span in queue_submit_spans]

    assert span_names.count("tests.integrations.django.test_tasks.task_one") == 2
    assert span_names.count("tests.integrations.django.test_tasks.task_two") == 1



from django.contrib.auth.models import User
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt


@csrf_exempt
def postgres_select_orm(request, *args, **kwargs):
    user = User.objects.using("postgres").all().first()
    return HttpResponse("ok {}".format(user))



import os
import threading
from unittest import mock

import kombu
import pytest
from celery import VERSION, Celery
from celery.bin import worker

import sentry_sdk
import sentry_sdk.traces
from sentry_sdk import get_current_span, start_transaction
from sentry_sdk.integrations.celery import (
    CeleryIntegration,
    _wrap_task_run,
)
from sentry_sdk.integrations.celery.beat import _get_headers
from sentry_sdk.utils import SENSITIVE_DATA_SUBSTITUTE
from tests.conftest import ApproxDict


@pytest.fixture
def connect_signal(request):
    def inner(signal, f):
        signal.connect(f)
        request.addfinalizer(lambda: signal.disconnect(f))

    return inner


@pytest.fixture
def init_celery(sentry_init, request):
    def inner(
        propagate_traces=True,
        backend="always_eager",
        monitor_beat_tasks=False,
        **kwargs,
    ):
        sentry_init(
            integrations=[
                CeleryIntegration(
                    propagate_traces=propagate_traces,
                    monitor_beat_tasks=monitor_beat_tasks,
                )
            ],
            **kwargs,
        )
        celery = Celery(__name__)

        if backend == "always_eager":
            if VERSION < (4,):
                celery.conf.CELERY_ALWAYS_EAGER = True
            else:
                celery.conf.task_always_eager = True
        elif backend == "redis":
            # broken on celery 3
            if VERSION < (4,):
                pytest.skip("Redis backend broken for some reason")

            # this backend requires capture_events_forksafe
            celery.conf.worker_max_tasks_per_child = 1
            celery.conf.worker_concurrency = 1
            redis_url = f"redis://{os.environ.get('SENTRY_PYTHON_TEST_REDIS_HOST', '127.0.0.1')}:6379"
            celery.conf.broker_url = redis_url
            celery.conf.result_backend = redis_url
            celery.conf.task_always_eager = False

            # Once we drop celery 3 we can use the celery_worker fixture
            if VERSION < (5,):
                worker_fn = worker.worker(app=celery).run
            else:
                from celery.bin.base import CLIContext

                worker_fn = lambda: worker.worker(
                    obj=CLIContext(app=celery, no_color=True, workdir=".", quiet=False),
                    args=[],
                )

            worker_thread = threading.Thread(target=worker_fn)
            worker_thread.daemon = True
            worker_thread.start()
        else:
            raise ValueError(backend)

        return celery

    return inner


@pytest.fixture
def celery(init_celery):
    return init_celery()


@pytest.fixture(
    params=[
        lambda task, x, y: (
            task.delay(x, y),
            {"args": [x, y], "kwargs": {}},
        ),
        lambda task, x, y: (
            task.apply_async((x, y)),
            {"args": [x, y], "kwargs": {}},
        ),
        lambda task, x, y: (
            task.apply_async(args=(x, y)),
            {"args": [x, y], "kwargs": {}},
        ),
        lambda task, x, y: (
            task.apply_async(kwargs=dict(x=x, y=y)),
            {"args": [], "kwargs": {"x": x, "y": y}},
        ),
    ]
)
def celery_invocation(request):
    """
    Invokes a task in multiple ways Celery allows you to (testing our apply_async monkeypatch).

    Currently limited to a task signature of the form foo(x, y)
    """
    return request.param


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("send_default_pii", [True, False])
def test_simple_with_performance(
    capture_events,
    capture_items,
    init_celery,
    celery_invocation,
    span_streaming,
    send_default_pii,
):
    celery = init_celery(
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task(name="dummy_task")
    def dummy_task(x, y):
        foo = 42  # noqa
        return x / y

    if span_streaming:
        items = capture_items("event", "span")

        with sentry_sdk.traces.start_span(name="span") as span:
            celery_invocation(dummy_task, 1, 2)
            _, expected_context = celery_invocation(dummy_task, 1, 0)

        sentry_sdk.flush()

        error_event = next(item.payload for item in items if item.type == "event")

        assert error_event["contexts"]["trace"]["trace_id"] == span.trace_id
        assert error_event["contexts"]["trace"]["span_id"] != span.span_id
    else:
        events = capture_events()

        with start_transaction(op="unit test transaction") as transaction:
            celery_invocation(dummy_task, 1, 2)
            _, expected_context = celery_invocation(dummy_task, 1, 0)

        (_, error_event, _, _) = events

        assert error_event["contexts"]["trace"]["trace_id"] == transaction.trace_id
        assert error_event["contexts"]["trace"]["span_id"] != transaction.span_id

    assert error_event["transaction"] == "dummy_task"
    assert "celery_task_id" in error_event["tags"]
    if send_default_pii:
        assert error_event["extra"]["celery-job"] == dict(
            task_name="dummy_task", **expected_context
        )
    else:
        assert error_event["extra"]["celery-job"] == {
            "task_name": "dummy_task",
            "args": SENSITIVE_DATA_SUBSTITUTE,
            "kwargs": SENSITIVE_DATA_SUBSTITUTE,
        }

    (exception,) = error_event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    assert exception["mechanism"]["type"] == "celery"
    assert exception["stacktrace"]["frames"][0]["vars"]["foo"] == "42"


@pytest.mark.parametrize("send_default_pii", [True, False])
def test_simple_without_performance(
    capture_events, init_celery, celery_invocation, send_default_pii
):
    celery = init_celery(traces_sample_rate=None, send_default_pii=send_default_pii)
    events = capture_events()

    @celery.task(name="dummy_task")
    def dummy_task(x, y):
        foo = 42  # noqa
        return x / y

    scope = sentry_sdk.get_isolation_scope()

    celery_invocation(dummy_task, 1, 2)
    _, expected_context = celery_invocation(dummy_task, 1, 0)

    (error_event,) = events

    assert (
        error_event["contexts"]["trace"]["trace_id"]
        == scope._propagation_context.trace_id
    )
    assert (
        error_event["contexts"]["trace"]["span_id"]
        != scope._propagation_context.span_id
    )
    assert error_event["transaction"] == "dummy_task"
    assert "celery_task_id" in error_event["tags"]
    if send_default_pii:
        assert error_event["extra"]["celery-job"] == dict(
            task_name="dummy_task", **expected_context
        )
    else:
        assert error_event["extra"]["celery-job"] == {
            "task_name": "dummy_task",
            "args": SENSITIVE_DATA_SUBSTITUTE,
            "kwargs": SENSITIVE_DATA_SUBSTITUTE,
        }

    (exception,) = error_event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"
    assert exception["mechanism"]["type"] == "celery"
    assert exception["stacktrace"]["frames"][0]["vars"]["foo"] == "42"


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("task_fails", [True, False], ids=["error", "success"])
def test_transaction_events(
    capture_events,
    capture_items,
    init_celery,
    celery_invocation,
    task_fails,
    span_streaming,
):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task(name="dummy_task")
    def dummy_task(x, y):
        return x / y

    # XXX: For some reason the first call does not get instrumented properly.
    celery_invocation(dummy_task, 1, 1)
    sentry_sdk.flush()

    if span_streaming:
        items = capture_items("event", "span")

        with sentry_sdk.traces.start_span(name="submission") as span:
            celery_invocation(dummy_task, 1, 0 if task_fails else 1)

        sentry_sdk.flush()

        if task_fails:
            error_event = items.pop(0).payload
            assert error_event["contexts"]["trace"]["trace_id"] == span.trace_id
            assert error_event["exception"]["values"][0]["type"] == "ZeroDivisionError"

        process_span, execution_span, submit_span, submission_span = [
            item.payload for item in items
        ]

        assert execution_span["name"] == "dummy_task"
        assert execution_span["is_segment"] is True
        assert execution_span["attributes"]["sentry.segment.name.source"] == "task"
        assert execution_span["trace_id"] == span.trace_id
        if task_fails:
            assert execution_span["status"] == "error"
        else:
            assert execution_span["status"] == "ok"

        assert process_span["name"] == "dummy_task"
        assert process_span["trace_id"] == span.trace_id
        assert process_span["attributes"]["sentry.op"] == "queue.process"
        assert process_span["parent_span_id"] == execution_span["span_id"]

        assert submission_span["name"] == "submission"
        assert submission_span["is_segment"] is True

        assert submit_span["name"] == "dummy_task"
        assert submit_span["attributes"]["sentry.op"] == "queue.submit.celery"
        assert submit_span["attributes"]["sentry.origin"] == "auto.queue.celery"
        assert (
            submit_span["parent_span_id"] == submission_span["span_id"] == span.span_id
        )
        assert submit_span["trace_id"] == span.trace_id

    else:
        events = capture_events()

        with start_transaction(name="submission") as transaction:
            celery_invocation(dummy_task, 1, 0 if task_fails else 1)

        if task_fails:
            error_event = events.pop(0)
            assert error_event["contexts"]["trace"]["trace_id"] == transaction.trace_id
            assert error_event["exception"]["values"][0]["type"] == "ZeroDivisionError"

        execution_event, submission_event = events
        assert execution_event["transaction"] == "dummy_task"
        assert execution_event["transaction_info"] == {"source": "task"}

        assert submission_event["transaction"] == "submission"
        assert submission_event["transaction_info"] == {"source": "custom"}

        assert execution_event["type"] == submission_event["type"] == "transaction"
        assert execution_event["contexts"]["trace"]["trace_id"] == transaction.trace_id
        assert submission_event["contexts"]["trace"]["trace_id"] == transaction.trace_id

        if task_fails:
            assert execution_event["contexts"]["trace"]["status"] == "internal_error"
        else:
            assert execution_event["contexts"]["trace"]["status"] == "ok"

        assert len(execution_event["spans"]) == 1
        assert (
            execution_event["spans"][0].items()
            >= {
                "trace_id": str(transaction.trace_id),
                "same_process_as_parent": True,
                "op": "queue.process",
                "description": "dummy_task",
                "data": ApproxDict(),
            }.items()
        )
        assert submission_event["spans"] == [
            {
                "data": ApproxDict(),
                "description": "dummy_task",
                "op": "queue.submit.celery",
                "origin": "auto.queue.celery",
                "parent_span_id": submission_event["contexts"]["trace"]["span_id"],
                "same_process_as_parent": True,
                "span_id": submission_event["spans"][0]["span_id"],
                "start_timestamp": submission_event["spans"][0]["start_timestamp"],
                "timestamp": submission_event["spans"][0]["timestamp"],
                "trace_id": str(transaction.trace_id),
            }
        ]


def test_no_double_patching(celery):
    """Ensure that Celery tasks are only patched once to prevent stack overflows.

    We used to have a bug in the Celery integration where its monkeypatching
    was repeated for every task invocation, leading to stackoverflows.

    See https://github.com/getsentry/sentry-python/issues/265
    """

    @celery.task(name="dummy_task")
    def dummy_task():
        return 42

    # Initially, the task should not be marked as patched
    assert not hasattr(dummy_task, "_sentry_is_patched")

    # First invocation should trigger patching
    result1 = dummy_task.delay()
    assert result1.get() == 42
    assert getattr(dummy_task, "_sentry_is_patched", False) is True

    patched_run = dummy_task.run

    # Second invocation should not re-patch
    result2 = dummy_task.delay()
    assert result2.get() == 42
    assert dummy_task.run is patched_run
    assert getattr(dummy_task, "_sentry_is_patched", False) is True


def test_simple_no_propagation(capture_events, init_celery):
    celery = init_celery(propagate_traces=False)
    events = capture_events()

    @celery.task(name="dummy_task")
    def dummy_task():
        1 / 0

    with start_transaction() as transaction:
        dummy_task.delay()

    (event,) = events
    assert event["contexts"]["trace"]["trace_id"] != transaction.trace_id
    assert event["transaction"] == "dummy_task"
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"


def test_ignore_expected(capture_events, celery):
    events = capture_events()

    @celery.task(name="dummy_task", throws=(ZeroDivisionError,))
    def dummy_task(x, y):
        return x / y

    dummy_task.delay(1, 2)
    dummy_task.delay(1, 0)
    assert not events


@pytest.mark.xfail(
    (4, 2, 0) <= VERSION < (4, 4, 3),
    strict=True,
    reason="https://github.com/celery/celery/issues/4661",
)
def test_retry(celery, capture_events):
    events = capture_events()
    failures = [True, True, False]
    runs = []

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self):
        runs.append(1)
        try:
            if failures.pop(0):
                1 / 0
        except Exception as exc:
            self.retry(max_retries=2, exc=exc)

    dummy_task.delay()

    assert len(runs) == 3
    assert not events

    failures = [True, True, True]
    runs = []

    dummy_task.delay()

    assert len(runs) == 3
    (event,) = events
    exceptions = event["exception"]["values"]

    for e in exceptions:
        assert e["type"] == "ZeroDivisionError"


@pytest.mark.skip(
    reason="This test is hanging when running test with `tox --parallel auto`. TODO: Figure out why and fix it!"
)
@pytest.mark.forked
def test_redis_backend_trace_propagation(init_celery, capture_events_forksafe):
    celery = init_celery(traces_sample_rate=1.0, backend="redis")

    events = capture_events_forksafe()

    runs = []

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self):
        runs.append(1)
        1 / 0

    with start_transaction(name="submit_celery"):
        # Curious: Cannot use delay() here or py2.7-celery-4.2 crashes
        res = dummy_task.apply_async()

    with pytest.raises(Exception):  # noqa: B017
        # Celery 4.1 raises a gibberish exception
        res.wait()

    # if this is nonempty, the worker never really forked
    assert not runs

    submit_transaction = events.read_event()
    assert submit_transaction["type"] == "transaction"
    assert submit_transaction["transaction"] == "submit_celery"

    assert len(submit_transaction["spans"]), (
        4
    )  # Because redis integration was auto enabled
    span = submit_transaction["spans"][0]
    assert span["op"] == "queue.submit.celery"
    assert span["description"] == "dummy_task"

    event = events.read_event()
    (exception,) = event["exception"]["values"]
    assert exception["type"] == "ZeroDivisionError"

    transaction = events.read_event()
    assert (
        transaction["contexts"]["trace"]["trace_id"]
        == event["contexts"]["trace"]["trace_id"]
        == submit_transaction["contexts"]["trace"]["trace_id"]
    )

    events.read_flush()

    # if this is nonempty, the worker never really forked
    assert not runs


@pytest.mark.forked
@pytest.mark.parametrize("newrelic_order", ["sentry_first", "sentry_last"])
def test_newrelic_interference(init_celery, newrelic_order, celery_invocation):
    def instrument_newrelic():
        try:
            # older newrelic versions
            import celery.app.trace as celery_trace_module
            from newrelic.hooks.application_celery import (
                instrument_celery_execute_trace,
            )

            assert hasattr(celery_trace_module, "build_tracer")
            instrument_celery_execute_trace(celery_trace_module)

        except ImportError:
            # newer newrelic versions
            import celery.app as celery_app_module
            from newrelic.hooks.application_celery import instrument_celery_app_base

            assert hasattr(celery_app_module, "Celery")
            assert hasattr(celery_app_module.Celery, "send_task")
            instrument_celery_app_base(celery_app_module)

    if newrelic_order == "sentry_first":
        celery = init_celery()
        instrument_newrelic()
    elif newrelic_order == "sentry_last":
        instrument_newrelic()
        celery = init_celery()
    else:
        raise ValueError(newrelic_order)

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self, x, y):
        return x / y

    assert dummy_task.apply(kwargs={"x": 1, "y": 1}).wait() == 1
    assert celery_invocation(dummy_task, 1, 1)[0].wait() == 1


@pytest.mark.parametrize("span_streaming", [True, False])
def test_traces_sampler_gets_task_info_in_sampling_context(
    span_streaming,
    init_celery,
    celery_invocation,
    DictionaryContaining,  # noqa:N803
):
    traces_sampler = mock.Mock(return_value=1.0)
    celery = init_celery(
        traces_sampler=traces_sampler,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task(name="dog_walk")
    def walk_dogs(x, y):
        dogs, route = x
        num_loops = y
        return dogs, route, num_loops

    _, args_kwargs = celery_invocation(
        walk_dogs, [["Maisey", "Charlie", "Bodhi", "Cory"], "Dog park round trip"], 1
    )

    traces_sampler.assert_any_call(
        # depending on the iteration of celery_invocation, the data might be
        # passed as args or as kwargs, so make this generic
        DictionaryContaining({"celery_job": dict(task="dog_walk", **args_kwargs)})
    )


def test_abstract_task(capture_events, celery, celery_invocation):
    events = capture_events()

    class AbstractTask(celery.Task):
        abstract = True

        def __call__(self, *args, **kwargs):
            try:
                return self.run(*args, **kwargs)
            except ZeroDivisionError:
                return None

    @celery.task(name="dummy_task", base=AbstractTask)
    def dummy_task(x, y):
        return x / y

    with start_transaction():
        celery_invocation(dummy_task, 1, 0)

    assert not events


def test_task_headers(celery):
    """
    Test that the headers set in the Celery Beat auto-instrumentation are passed to the celery signal handlers
    """
    sentry_crons_setup = {
        "sentry-monitor-slug": "some-slug",
        "sentry-monitor-config": {"some": "config"},
        "sentry-monitor-check-in-id": "123abc",
    }

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self, x, y):
        return _get_headers(self)

    # This is how the Celery Beat auto-instrumentation starts a task
    # in the monkey patched version of `apply_async`
    # in `sentry_sdk/integrations/celery.py::_wrap_apply_async()`
    result = dummy_task.apply_async(args=(1, 0), headers=sentry_crons_setup)

    expected_headers = sentry_crons_setup.copy()
    # Newly added headers
    expected_headers["sentry-trace"] = mock.ANY
    expected_headers["baggage"] = mock.ANY
    expected_headers["sentry-task-enqueued-time"] = mock.ANY

    assert result.get() == expected_headers


def test_baggage_propagation(init_celery):
    celery = init_celery(traces_sample_rate=1.0, release="abcdef")

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self, x, y):
        return _get_headers(self)

    # patch random.randrange to return a predictable sample_rand value
    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=500000):
        with start_transaction() as transaction:
            result = dummy_task.apply_async(
                args=(1, 0),
                headers={"baggage": "custom=value"},
            ).get()

            assert sorted(result["baggage"].split(",")) == sorted(
                [
                    "sentry-release=abcdef",
                    "sentry-trace_id={}".format(transaction.trace_id),
                    "sentry-environment=production",
                    "sentry-sample_rand=0.500000",
                    "sentry-sample_rate=1.0",
                    "sentry-sampled=true",
                    "custom=value",
                ]
            )


@pytest.mark.parametrize("span_streaming", [True, False])
def test_sentry_propagate_traces_override(span_streaming, init_celery):
    """
    Test if the `sentry-propagate-traces` header given to `apply_async`
    overrides the `propagate_traces` parameter in the integration constructor.
    """
    celery = init_celery(
        propagate_traces=True,
        traces_sample_rate=1.0,
        release="abcdef",
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task(name="dummy_task", bind=True)
    def dummy_task(self, message):
        trace_id = (
            sentry_sdk.traces.get_current_span().trace_id
            if span_streaming
            else get_current_span().trace_id
        )
        return trace_id

    if span_streaming:
        with sentry_sdk.traces.start_span(name="parent") as span:
            parent_trace_id = span.trace_id

            # should propagate trace
            task_trace_id = dummy_task.apply_async(
                args=("some message",),
            ).get()
            assert parent_trace_id == task_trace_id

            # should NOT propagate trace
            task_trace_id = dummy_task.apply_async(
                args=("another message",),
                headers={"sentry-propagate-traces": False},
            ).get()
            assert parent_trace_id != task_trace_id
    else:
        with start_transaction() as transaction:
            transaction_trace_id = transaction.trace_id

            # should propagate trace
            task_trace_id = dummy_task.apply_async(
                args=("some message",),
            ).get()
            assert transaction_trace_id == task_trace_id

            # should NOT propagate trace
            task_trace_id = dummy_task.apply_async(
                args=("another message",),
                headers={"sentry-propagate-traces": False},
            ).get()
            assert transaction_trace_id != task_trace_id


def test_apply_async_manually_span(sentry_init):
    sentry_init(
        integrations=[CeleryIntegration()],
    )

    def dummy_function(*args, **kwargs):
        headers = kwargs.get("headers")
        assert "sentry-trace" in headers
        assert "baggage" in headers

    wrapped = _wrap_task_run(dummy_function)
    wrapped(mock.MagicMock(), (), headers={})


def test_apply_async_no_args(init_celery):
    celery = init_celery()

    @celery.task
    def example_task():
        return "success"

    try:
        result = example_task.apply_async(None, {})
    except TypeError:
        pytest.fail("Calling `apply_async` without arguments raised a TypeError")

    assert result.get() == "success"


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("routing_key", ("celery", "custom"))
@mock.patch("celery.app.task.Task.request")
def test_messaging_destination_name_default_exchange(
    mock_request,
    routing_key,
    span_streaming,
    init_celery,
    capture_events,
    capture_items,
):
    celery_app = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    mock_request.delivery_info = {"routing_key": routing_key, "exchange": ""}

    @celery_app.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert (
            process_span.payload["attributes"]["messaging.destination.name"]
            == routing_key
        )
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert span["data"]["messaging.destination.name"] == routing_key


@pytest.mark.parametrize("span_streaming", [True, False])
@mock.patch("celery.app.task.Task.request")
def test_messaging_destination_name_nondefault_exchange(
    mock_request, span_streaming, init_celery, capture_events, capture_items
):
    """
    Currently, we only capture the routing key as the messaging.destination.name when
    we are using the default exchange (""). This is because the default exchange ensures
    that the routing key is the queue name. Other exchanges may not guarantee this
    behavior.
    """
    celery_app = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    mock_request.delivery_info = {"routing_key": "celery", "exchange": "custom"}

    @celery_app.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert "messaging.destination.name" not in process_span.payload["attributes"]
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert "messaging.destination.name" not in span["data"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_messaging_id(span_streaming, init_celery, capture_events, capture_items):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task
    def example_task(): ...

    if span_streaming:
        items = capture_items("span")
        example_task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert "messaging.message.id" in process_span.payload["attributes"]
    else:
        events = capture_events()
        example_task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert "messaging.message.id" in span["data"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_retry_count_zero(span_streaming, init_celery, capture_events, capture_items):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert process_span.payload["attributes"]["messaging.message.retry.count"] == 0
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert span["data"]["messaging.message.retry.count"] == 0


@pytest.mark.parametrize("span_streaming", [True, False])
@mock.patch("celery.app.task.Task.request")
def test_retry_count_nonzero(
    mock_request, span_streaming, init_celery, capture_events, capture_items
):
    mock_request.retries = 3

    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert process_span.payload["attributes"]["messaging.message.retry.count"] == 3
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert span["data"]["messaging.message.retry.count"] == 3


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("system", ("redis", "amqp"))
def test_messaging_system(
    system, span_streaming, init_celery, capture_events, capture_items
):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    # Does not need to be a real URL, since we use always eager
    celery.conf.broker_url = f"{system}://example.com"  # noqa: E231

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert process_span.payload["attributes"]["messaging.system"] == system
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert span["data"]["messaging.system"] == system


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.parametrize("system", ("amqp", "redis"))
def test_producer_span_data(
    system, span_streaming, monkeypatch, sentry_init, capture_events, capture_items
):
    old_publish = kombu.messaging.Producer._publish

    def publish(*args, **kwargs):
        pass

    monkeypatch.setattr(kombu.messaging.Producer, "_publish", publish)

    sentry_init(
        integrations=[CeleryIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    celery = Celery(__name__, broker=f"{system}://example.com")  # noqa: E231

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="producer test"):
            task.apply_async()

        sentry_sdk.flush()

        span_items = [item.payload for item in items]
        publish_span = next(
            s for s in span_items if s["attributes"].get("sentry.op") == "queue.publish"
        )

        assert publish_span["attributes"]["messaging.system"] == system
        assert publish_span["attributes"]["messaging.destination.name"] == "celery"
        assert "messaging.message.id" in publish_span["attributes"]
        assert publish_span["attributes"]["messaging.message.retry.count"] == 0
    else:
        events = capture_events()

        with start_transaction():
            task.apply_async()

        (event,) = events
        span = next(span for span in event["spans"] if span["op"] == "queue.publish")

        assert span["data"]["messaging.system"] == system
        assert span["data"]["messaging.destination.name"] == "celery"
        assert "messaging.message.id" in span["data"]
        assert span["data"]["messaging.message.retry.count"] == 0

    monkeypatch.setattr(kombu.messaging.Producer, "_publish", old_publish)


@pytest.mark.parametrize("span_streaming", [True, False])
def test_receive_latency(span_streaming, init_celery, capture_events, capture_items):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, _execution_span = items
        assert "messaging.message.receive.latency" in process_span.payload["attributes"]
        assert (
            process_span.payload["attributes"]["messaging.message.receive.latency"] > 0
        )
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        (span,) = event["spans"]
        assert "messaging.message.receive.latency" in span["data"]
        assert span["data"]["messaging.message.receive.latency"] > 0


@pytest.mark.parametrize("span_streaming", [True, False])
def tests_span_origin_consumer(
    span_streaming, init_celery, capture_events, capture_items
):
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    celery.conf.broker_url = "redis://example.com"  # noqa: E231

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")
        task.apply_async()
        sentry_sdk.flush()
        process_span, execution_span = items
        assert (
            execution_span.payload["attributes"]["sentry.origin"] == "auto.queue.celery"
        )
        assert (
            process_span.payload["attributes"]["sentry.origin"] == "auto.queue.celery"
        )
    else:
        events = capture_events()
        task.apply_async()
        (event,) = events
        assert event["contexts"]["trace"]["origin"] == "auto.queue.celery"
        assert event["spans"][0]["origin"] == "auto.queue.celery"


@pytest.mark.parametrize("span_streaming", [True, False])
def tests_span_origin_producer(
    span_streaming, monkeypatch, sentry_init, capture_events, capture_items
):
    old_publish = kombu.messaging.Producer._publish

    def publish(*args, **kwargs):
        pass

    monkeypatch.setattr(kombu.messaging.Producer, "_publish", publish)

    sentry_init(
        integrations=[CeleryIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    celery = Celery(__name__, broker="redis://example.com")  # noqa: E231

    @celery.task()
    def task(): ...

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"):
            task.apply_async()

        sentry_sdk.flush()

        parent = items.pop(-1).payload
        assert parent["name"] == "custom parent"
        assert parent["attributes"]["sentry.origin"] == "manual"

        for item in items:
            assert item.payload["attributes"]["sentry.origin"] == "auto.queue.celery"
    else:
        events = capture_events()

        with start_transaction(name="custom_transaction"):
            task.apply_async()

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"

        for span in event["spans"]:
            assert span["origin"] == "auto.queue.celery"

    monkeypatch.setattr(kombu.messaging.Producer, "_publish", old_publish)


@pytest.mark.parametrize("span_streaming", [True, False])
@pytest.mark.forked
@mock.patch("celery.Celery.send_task")
def test_send_task_wrapped(
    patched_send_task,
    span_streaming,
    sentry_init,
    capture_events,
    capture_items,
    reset_integrations,
):
    sentry_init(
        integrations=[CeleryIntegration()],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    celery = Celery(__name__, broker="redis://example.com")  # noqa: E231

    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent") as outer_span:
            celery.send_task(
                "very_creative_task_name", args=(1, 2), kwargs={"foo": "bar"}
            )
        sentry_sdk.flush()
    else:
        events = capture_events()
        with sentry_sdk.start_transaction(name="custom_transaction"):
            celery.send_task(
                "very_creative_task_name", args=(1, 2), kwargs={"foo": "bar"}
            )

    (call,) = patched_send_task.call_args_list  # We should have exactly one call
    (args, kwargs) = call

    assert args == (celery, "very_creative_task_name")
    assert kwargs["args"] == (1, 2)
    assert kwargs["kwargs"] == {"foo": "bar"}
    assert set(kwargs["headers"].keys()) == {
        "sentry-task-enqueued-time",
        "sentry-trace",
        "baggage",
        "headers",
    }
    assert set(kwargs["headers"]["headers"].keys()) == {
        "sentry-trace",
        "baggage",
        "sentry-task-enqueued-time",
    }
    assert (
        kwargs["headers"]["sentry-trace"]
        == kwargs["headers"]["headers"]["sentry-trace"]
    )

    if span_streaming:
        submit_span, outer = [item.payload for item in items]

        assert outer["name"] == "custom parent"
        assert outer["is_segment"] is True

        assert submit_span["name"] == "very_creative_task_name"
        assert submit_span["attributes"]["sentry.op"] == "queue.submit.celery"
        assert submit_span["trace_id"] == outer_span.trace_id
        assert (
            submit_span["trace_id"] == kwargs["headers"]["sentry-trace"].split("-")[0]
        )

    else:
        (event,) = events
        assert event["type"] == "transaction"
        assert event["transaction"] == "custom_transaction"

        (span,) = event["spans"]
        assert span["description"] == "very_creative_task_name"
        assert span["op"] == "queue.submit.celery"
        assert span["trace_id"] == kwargs["headers"]["sentry-trace"].split("-")[0]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_user_custom_headers_accessible_in_task(span_streaming, init_celery):
    """
    Regression test for https://github.com/getsentry/sentry-python/issues/5566

    User-provided custom headers passed to apply_async() must be accessible
    via task.request.headers on the worker side.
    """
    celery = init_celery(
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    @celery.task(name="custom_headers_task", bind=True)
    def custom_headers_task(self):
        return dict(self.request.headers or {})

    custom_headers = {
        "my_custom_key": "my_value",
        "correlation_id": "abc-123",
        "tenant_id": "tenant-42",
    }

    if span_streaming:
        with sentry_sdk.traces.start_span(name="test"):
            result = custom_headers_task.apply_async(headers=custom_headers)
    else:
        with start_transaction(name="test"):
            result = custom_headers_task.apply_async(headers=custom_headers)

    received_headers = result.get()
    for key, value in custom_headers.items():
        assert received_headers.get(key) == value, (
            f"Custom header {key!r} not found in task.request.headers"
        )


@pytest.mark.skip(reason="placeholder so that forked test does not come last")
def test_placeholder():
    """Forked tests must not come last in the module.
    See https://github.com/pytest-dev/pytest-forked/issues/67#issuecomment-1964718720.
    """
    pass



import time
from typing import TYPE_CHECKING, Any, Optional

import sentry_sdk
from sentry_sdk.utils import format_attribute

if TYPE_CHECKING:
    from sentry_sdk._types import Attributes, Metric, MetricType


def _capture_metric(
    name: str,
    metric_type: "MetricType",
    value: float,
    unit: "Optional[str]" = None,
    attributes: "Optional[Attributes]" = None,
) -> None:
    attrs: "Attributes" = {}

    if attributes:
        for k, v in attributes.items():
            attrs[k] = format_attribute(v)

    metric: "Metric" = {
        "timestamp": time.time(),
        "trace_id": None,
        "span_id": None,
        "name": name,
        "type": metric_type,
        "value": float(value),
        "unit": unit,
        "attributes": attrs,
    }

    sentry_sdk.get_current_scope()._capture_metric(metric)


def count(
    name: str,
    value: float,
    unit: "Optional[str]" = None,
    attributes: "Optional[dict[str, Any]]" = None,
) -> None:
    _capture_metric(name, "counter", value, unit, attributes)


def gauge(
    name: str,
    value: float,
    unit: "Optional[str]" = None,
    attributes: "Optional[dict[str, Any]]" = None,
) -> None:
    _capture_metric(name, "gauge", value, unit, attributes)


def distribution(
    name: str,
    value: float,
    unit: "Optional[str]" = None,
    attributes: "Optional[dict[str, Any]]" = None,
) -> None:
    _capture_metric(name, "distribution", value, unit, attributes)



import os
import sys
from collections import Counter
from unittest import mock

import pytest

import sentry_sdk
from sentry_sdk.transport import Transport


class HealthyTestTransport(Transport):
    def capture_envelope(self, _):
        pass

    def is_healthy(self):
        return True


class UnhealthyTestTransport(HealthyTestTransport):
    def is_healthy(self):
        return False


def test_no_monitor_if_disabled(sentry_init):
    sentry_init(
        transport=HealthyTestTransport(),
        enable_backpressure_handling=False,
    )

    assert sentry_sdk.get_client().monitor is None


def test_monitor_if_enabled(sentry_init):
    sentry_init(transport=HealthyTestTransport())

    monitor = sentry_sdk.get_client().monitor
    assert monitor is not None
    assert monitor._thread is None

    assert monitor.is_healthy() is True
    assert monitor.downsample_factor == 0
    assert monitor._thread is not None
    assert monitor._thread.name == "sentry.monitor"


def test_monitor_unhealthy(sentry_init):
    sentry_init(transport=UnhealthyTestTransport())

    monitor = sentry_sdk.get_client().monitor
    monitor.interval = 0.1

    assert monitor.is_healthy() is True

    for i in range(15):
        monitor.run()
        assert monitor.is_healthy() is False
        assert monitor.downsample_factor == (i + 1 if i < 10 else 10)


def test_transaction_uses_downsampled_rate(
    sentry_init, capture_record_lost_event_calls, monkeypatch
):
    sentry_init(
        traces_sample_rate=1.0,
        transport=UnhealthyTestTransport(),
    )

    record_lost_event_calls = capture_record_lost_event_calls()

    monitor = sentry_sdk.get_client().monitor
    monitor.interval = 0.1

    assert monitor.is_healthy() is True
    monitor.run()
    assert monitor.is_healthy() is False
    assert monitor.downsample_factor == 1

    # make sure we don't sample the transaction
    with mock.patch("sentry_sdk.tracing_utils.Random.randrange", return_value=750000):
        with sentry_sdk.start_transaction(name="foobar") as transaction:
            assert transaction.sampled is False
            assert transaction.sample_rate == 0.5

    assert Counter(record_lost_event_calls) == Counter(
        [
            ("backpressure", "transaction", None, 1),
            ("backpressure", "span", None, 1),
        ]
    )


def test_monitor_no_thread_on_shutdown_no_errors(sentry_init):
    sentry_init(transport=HealthyTestTransport())

    # make it seem like the interpreter is shutting down
    with mock.patch(
        "threading.Thread.start",
        side_effect=RuntimeError("can't create new thread at interpreter shutdown"),
    ):
        monitor = sentry_sdk.get_client().monitor
        assert monitor is not None
        assert monitor._thread is None
        monitor.run()
        assert monitor._thread is None


@pytest.mark.skipif(
    sys.platform == "win32"
    or not hasattr(os, "fork")
    or not hasattr(os, "register_at_fork"),
    reason="requires POSIX fork and os.register_at_fork (Python 3.7+)",
)
def test_monitor_thread_lock_reset_in_child_after_fork(sentry_init):
    """Regression test for #6148.

    If os.fork() runs while Monitor._thread_lock is held, the child
    inherits the lock locked. The holding thread does not exist in the
    child, so the lock can never be released and _ensure_running
    deadlocks forever. The after-fork hook must replace the lock with
    a fresh one in the child.
    """
    sentry_init(transport=HealthyTestTransport())
    monitor = sentry_sdk.get_client().monitor
    assert monitor is not None

    original_lock = monitor._thread_lock
    original_lock.acquire()
    pid = os.fork()
    if pid == 0:
        # Child: was the lock object replaced and is the new one not
        # held? Without the fix, _thread_lock is `original_lock`
        # inherited locked, so `replaced` is False. blocking=False
        # guarantees the child can't hang on a regression.
        replaced = monitor._thread_lock is not original_lock
        unheld = monitor._thread_lock.acquire(blocking=False)
        os._exit(0 if replaced and unheld else 1)

    original_lock.release()
    _, status = os.waitpid(pid, 0)
    assert os.WIFEXITED(status) and os.WEXITSTATUS(status) == 0



import os

import sentry_sdk
from sentry_sdk.integrations.aws_lambda import AwsLambdaIntegration

sentry_sdk.init(
    dsn=os.environ.get("SENTRY_DSN"),
    traces_sample_rate=1.0,
    integrations=[AwsLambdaIntegration()],
    trace_lifecycle="stream",
)


def handler(event, context):
    return {"event": event}



from unittest import mock

import boto3
import pytest

import sentry_sdk
from sentry_sdk.integrations.boto3 import Boto3Integration
from tests.conftest import ApproxDict
from tests.integrations.boto3 import read_fixture
from tests.integrations.boto3.aws_mock import MockResponse

session = boto3.Session(
    aws_access_key_id="-",
    aws_secret_access_key="-",
)


@pytest.mark.parametrize("span_streaming", [True, False])
def test_basic(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Boto3Integration()],
        # disabled because session.resource() or s3.Bucket() result in a subprocess span for a
        # shell that runs "uname -p 2> /dev/null" on Python 3.7 with boto3 version 1.12.49.
        default_integrations=False,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket("bucket")

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent") as span, MockResponse(
            s3.meta.client, 200, {}, read_fixture("s3_list.xml")
        ):
            objects = [obj for obj in bucket.objects.all()]
            assert len(objects) == 2
            assert objects[0].key == "foo.txt"
            assert objects[1].key == "bar.txt"
            span.end()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "http.client"
        assert span["name"] == "aws.s3.ListObjects"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction() as transaction, MockResponse(
            s3.meta.client, 200, {}, read_fixture("s3_list.xml")
        ):
            items = [obj for obj in bucket.objects.all()]
            assert len(items) == 2
            assert items[0].key == "foo.txt"
            assert items[1].key == "bar.txt"
            transaction.finish()

        (event,) = events
        assert event["type"] == "transaction"
        assert len(event["spans"]) == 1
        (span,) = event["spans"]
        assert span["op"] == "http.client"
        assert span["description"] == "aws.s3.ListObjects"


@pytest.mark.parametrize("send_default_pii", [True, False])
@pytest.mark.parametrize("span_streaming", [True, False])
def test_streaming(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
    send_default_pii,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Boto3Integration()],
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    s3 = session.resource("s3")
    obj = s3.Bucket("bucket").Object("foo.pdf")

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent") as span, MockResponse(
            s3.meta.client, 200, {}, b"hello"
        ):
            body = obj.get()["Body"]
            assert body.read(1) == b"h"
            assert body.read(2) == b"el"
            assert body.read(3) == b"lo"
            assert body.read(1) == b""
            span.end()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 3

        span1 = spans[0]
        assert span1["attributes"]["sentry.op"] == "http.client"
        assert span1["name"] == "aws.s3.GetObject"

        expected_attrs = {
            "http.request.method": "GET",
            "rpc.method": "S3/GetObject",
            "sentry.environment": "production",
            "sentry.op": "http.client",
            "sentry.origin": "auto.http.boto3",
            "sentry.release": mock.ANY,
            "sentry.sdk.name": "sentry.python",
            "sentry.sdk.version": mock.ANY,
            "sentry.segment.id": mock.ANY,
            "sentry.segment.name": "custom parent",
            "server.address": mock.ANY,
            "thread.id": mock.ANY,
            "thread.name": mock.ANY,
        }
        if send_default_pii:
            expected_attrs["url.full"] = "https://bucket.s3.amazonaws.com/foo.pdf"
            expected_attrs["url.fragment"] = ""
            expected_attrs["url.query"] = ""
        assert span1["attributes"] == ApproxDict(expected_attrs)

        if not send_default_pii:
            assert "url.full" not in span1["attributes"]
            assert "url.fragment" not in span1["attributes"]
            assert "url.query" not in span1["attributes"]

        span2 = spans[1]
        assert span2["attributes"]["sentry.op"] == "http.client.stream"
        assert span2["name"] == "aws.s3.GetObject"
        assert span2["parent_span_id"] == span1["span_id"]
    else:
        events = capture_events()

        with sentry_sdk.start_transaction() as transaction, MockResponse(
            s3.meta.client, 200, {}, b"hello"
        ):
            body = obj.get()["Body"]
            assert body.read(1) == b"h"
            assert body.read(2) == b"el"
            assert body.read(3) == b"lo"
            assert body.read(1) == b""
            transaction.finish()

        (event,) = events
        assert event["type"] == "transaction"
        assert len(event["spans"]) == 2

        span1 = event["spans"][0]
        assert span1["op"] == "http.client"
        assert span1["description"] == "aws.s3.GetObject"
        assert span1["data"] == ApproxDict(
            {
                "http.method": "GET",
                "aws.request.url": "https://bucket.s3.amazonaws.com/foo.pdf",
                "http.fragment": "",
                "http.query": "",
            }
        )

        span2 = event["spans"][1]
        assert span2["op"] == "http.client.stream"
        assert span2["description"] == "aws.s3.GetObject"
        assert span2["parent_span_id"] == span1["span_id"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_streaming_close(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Boto3Integration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    s3 = session.resource("s3")
    obj = s3.Bucket("bucket").Object("foo.pdf")

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent") as span, MockResponse(
            s3.meta.client, 200, {}, b"hello"
        ):
            body = obj.get()["Body"]
            assert body.read(1) == b"h"
            body.close()  # close partially-read stream
            span.end()

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 3
        span1 = spans[0]
        assert span1["attributes"]["sentry.op"] == "http.client"
        span2 = spans[1]
        assert span2["attributes"]["sentry.op"] == "http.client.stream"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction() as transaction, MockResponse(
            s3.meta.client, 200, {}, b"hello"
        ):
            body = obj.get()["Body"]
            assert body.read(1) == b"h"
            body.close()  # close partially-read stream
            transaction.finish()

        (event,) = events
        assert event["type"] == "transaction"
        assert len(event["spans"]) == 2
        span1 = event["spans"][0]
        assert span1["op"] == "http.client"
        span2 = event["spans"][1]
        assert span2["op"] == "http.client.stream"


@pytest.mark.tests_internal_exceptions
@pytest.mark.parametrize("span_streaming", [True, False])
def test_omit_url_data_if_parsing_fails(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Boto3Integration()],
        send_default_pii=True,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket("bucket")

    if span_streaming:
        items = capture_items("span")

        with mock.patch(
            "sentry_sdk.integrations.boto3.parse_url",
            side_effect=ValueError,
        ):
            with sentry_sdk.traces.start_span(
                name="custom parent"
            ) as span, MockResponse(
                s3.meta.client, 200, {}, read_fixture("s3_list.xml")
            ):
                objects = [obj for obj in bucket.objects.all()]
                assert len(objects) == 2
                assert objects[0].key == "foo.txt"
                assert objects[1].key == "bar.txt"
                span.end()

                sentry_sdk.flush()
                spans = [item.payload for item in items]
                assert spans[0]["attributes"] == ApproxDict(
                    {
                        "http.request.method": "GET",
                        "rpc.method": "S3/ListObjects",
                        "sentry.environment": "production",
                        "sentry.op": "http.client",
                        "sentry.origin": "auto.http.boto3",
                        "sentry.release": mock.ANY,
                        "sentry.sdk.name": "sentry.python",
                        "sentry.sdk.version": mock.ANY,
                        "sentry.segment.id": mock.ANY,
                        "sentry.segment.name": "custom parent",
                        "server.address": mock.ANY,
                        "thread.id": mock.ANY,
                        "thread.name": mock.ANY,
                    }
                )

        assert "url.full" not in spans[0]["attributes"]
        assert "url.fragment" not in spans[0]["attributes"]
        assert "url.query" not in spans[0]["attributes"]
    else:
        events = capture_events()

        with mock.patch(
            "sentry_sdk.integrations.boto3.parse_url",
            side_effect=ValueError,
        ):
            with sentry_sdk.start_transaction() as transaction, MockResponse(
                s3.meta.client, 200, {}, read_fixture("s3_list.xml")
            ):
                items = [obj for obj in bucket.objects.all()]
                assert len(items) == 2
                assert items[0].key == "foo.txt"
                assert items[1].key == "bar.txt"
                transaction.finish()

                (event,) = events
                assert event["spans"][0]["data"] == ApproxDict(
                    {
                        "http.method": "GET",
                        # no url data
                    }
                )

        assert "aws.request.url" not in event["spans"][0]["data"]
        assert "http.fragment" not in event["spans"][0]["data"]
        assert "http.query" not in event["spans"][0]["data"]


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_origin(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    sentry_init(
        traces_sample_rate=1.0,
        integrations=[Boto3Integration()],
        trace_lifecycle="stream" if span_streaming else "static",
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket("bucket")

    if span_streaming:
        items = capture_items("span")

        with sentry_sdk.traces.start_span(name="custom parent"), MockResponse(
            s3.meta.client, 200, {}, read_fixture("s3_list.xml")
        ):
            _ = [obj for obj in bucket.objects.all()]

        sentry_sdk.flush()
        spans = [item.payload for item in items]

        assert spans[1]["attributes"]["sentry.origin"] == "manual"
        assert spans[0]["attributes"]["sentry.origin"] == "auto.http.boto3"
    else:
        events = capture_events()

        with sentry_sdk.start_transaction(), MockResponse(
            s3.meta.client, 200, {}, read_fixture("s3_list.xml")
        ):
            _ = [obj for obj in bucket.objects.all()]

        (event,) = events

        assert event["contexts"]["trace"]["origin"] == "manual"
        assert event["spans"][0]["origin"] == "auto.http.boto3"



import subprocess
import sys
from textwrap import dedent

import pytest

TEST_PARAMETERS = [("", "HttpTransport")]

if sys.version_info >= (3, 8):
    TEST_PARAMETERS.append(('_experiments={"transport_http2": True}', "Http2Transport"))


@pytest.mark.parametrize("options, transport", TEST_PARAMETERS)
def test_excepthook(tmpdir, options, transport):
    app = tmpdir.join("app.py")
    app.write(
        dedent(
            """
    from sentry_sdk import init, transport

    def capture_envelope(self, envelope):
        print("capture_envelope was called")
        event = envelope.get_event()
        if event is not None:
            print(event)

    transport.{transport}.capture_envelope = capture_envelope

    init("http://foobar@localhost/123", {options})

    frame_value = "LOL"

    1/0
    """.format(transport=transport, options=options)
        )
    )

    with pytest.raises(subprocess.CalledProcessError) as excinfo:
        subprocess.check_output([sys.executable, str(app)], stderr=subprocess.STDOUT)

    output = excinfo.value.output

    assert b"ZeroDivisionError" in output
    assert b"LOL" in output
    assert b"capture_envelope was called" in output


@pytest.mark.parametrize("options, transport", TEST_PARAMETERS)
def test_always_value_excepthook(tmpdir, options, transport):
    app = tmpdir.join("app.py")
    app.write(
        dedent(
            """
    import sys
    from sentry_sdk import init, transport
    from sentry_sdk.integrations.excepthook import ExcepthookIntegration

    def capture_envelope(self, envelope):
        print("capture_envelope was called")
        event = envelope.get_event()
        if event is not None:
            print(event)

    transport.{transport}.capture_envelope = capture_envelope

    sys.ps1 = "always_value_test"
    init("http://foobar@localhost/123",
        integrations=[ExcepthookIntegration(always_run=True)],
        {options}
    )

    frame_value = "LOL"

    1/0
    """.format(transport=transport, options=options)
        )
    )

    with pytest.raises(subprocess.CalledProcessError) as excinfo:
        subprocess.check_output([sys.executable, str(app)], stderr=subprocess.STDOUT)

    output = excinfo.value.output

    assert b"ZeroDivisionError" in output
    assert b"LOL" in output
    assert b"capture_envelope was called" in output



from string import Template
from typing import Dict

import pytest

import sentry_sdk
from sentry_sdk import capture_message, start_transaction
from sentry_sdk.integrations.rust_tracing import (
    EventTypeMapping,
    RustTracingIntegration,
    RustTracingLayer,
    RustTracingLevel,
)


def _test_event_type_mapping(metadata: Dict[str, object]) -> EventTypeMapping:
    level = RustTracingLevel(metadata.get("level"))
    if level == RustTracingLevel.Error:
        return EventTypeMapping.Exc
    elif level in (RustTracingLevel.Warn, RustTracingLevel.Info):
        return EventTypeMapping.Breadcrumb
    elif level == RustTracingLevel.Debug:
        return EventTypeMapping.Event
    elif level == RustTracingLevel.Trace:
        return EventTypeMapping.Ignore
    else:
        return EventTypeMapping.Ignore


class FakeRustTracing:
    # Parameters: `level`, `index`
    span_template = Template(
        """{"index":$index,"is_root":false,"metadata":{"fields":["index","use_memoized","version"],"file":"src/lib.rs","is_event":false,"is_span":true,"level":"$level","line":40,"module_path":"_bindings","name":"fibonacci","target":"_bindings"},"parent":null,"use_memoized":true}"""
    )

    # Parameters: `level`, `index`
    event_template = Template(
        """{"message":"Getting the ${index}th fibonacci number","metadata":{"fields":["message"],"file":"src/lib.rs","is_event":true,"is_span":false,"level":"$level","line":23,"module_path":"_bindings","name":"event src/lib.rs:23","target":"_bindings"}}"""
    )

    def __init__(self):
        self.spans = {}

    def set_layer_impl(self, layer: RustTracingLayer):
        self.layer = layer

    def new_span(self, level: RustTracingLevel, span_id: int, index_arg: int = 10):
        span_attrs = self.span_template.substitute(level=level.value, index=index_arg)
        state = self.layer.on_new_span(span_attrs, str(span_id))
        self.spans[span_id] = state

    def close_span(self, span_id: int):
        state = self.spans.pop(span_id)
        self.layer.on_close(str(span_id), state)

    def event(self, level: RustTracingLevel, span_id: int, index_arg: int = 10):
        event = self.event_template.substitute(level=level.value, index=index_arg)
        state = self.spans[span_id]
        self.layer.on_event(event, state)

    def record(self, span_id: int):
        state = self.spans[span_id]
        self.layer.on_record(str(span_id), """{"version": "memoized"}""", state)


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_new_span_on_close(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_new_span_on_close",
        initializer=rust_tracing.set_layer_impl,
        include_tracing_fields=True,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            sentry_first_rust_span = sentry_sdk.traces.get_current_span()
            rust_first_rust_span = rust_tracing.spans[3]

            assert sentry_first_rust_span == rust_first_rust_span

            rust_tracing.close_span(3)
            assert sentry_sdk.traces.get_current_span() != sentry_first_rust_span

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2

        # Ensure the span metadata is wired up
        span = spans[0]
        assert span["attributes"]["sentry.op"] == "function"
        assert (
            span["attributes"]["sentry.origin"]
            == "auto.function.rust_tracing.test_on_new_span_on_close"
        )
        assert span["name"] == "_bindings::fibonacci"

        # Ensure the span was opened/closed appropriately
        assert span["start_timestamp"] is not None
        assert span["end_timestamp"] is not None

        # Ensure the extra data from Rust is hooked up
        data = span["attributes"]
        assert data["use_memoized"]
        assert data["index"] == 10
        assert data["version"] == "None"
    else:
        events = capture_events()
        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            sentry_first_rust_span = sentry_sdk.get_current_span()
            rust_first_rust_span = rust_tracing.spans[3]

            assert sentry_first_rust_span == rust_first_rust_span

            rust_tracing.close_span(3)
            assert sentry_sdk.get_current_span() != sentry_first_rust_span

        (event,) = events
        assert len(event["spans"]) == 1

        # Ensure the span metadata is wired up
        span = event["spans"][0]
        assert span["op"] == "function"
        assert span["origin"] == "auto.function.rust_tracing.test_on_new_span_on_close"
        assert span["description"] == "_bindings::fibonacci"

        # Ensure the span was opened/closed appropriately
        assert span["start_timestamp"] is not None
        assert span["timestamp"] is not None

        # Ensure the extra data from Rust is hooked up
        data = span["data"]
        assert data["use_memoized"]
        assert data["index"] == 10
        assert data["version"] is None


@pytest.mark.parametrize("span_streaming", [True, False])
def test_nested_on_new_span_on_close(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_nested_on_new_span_on_close",
        initializer=rust_tracing.set_layer_impl,
        include_tracing_fields=True,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            original_sentry_span = sentry_sdk.traces.get_current_span()

            rust_tracing.new_span(RustTracingLevel.Info, 3, index_arg=10)
            sentry_first_rust_span = sentry_sdk.traces.get_current_span()
            rust_first_rust_span = rust_tracing.spans[3]

            # Use a different `index_arg` value for the inner span to help
            # distinguish the two at the end of the test
            rust_tracing.new_span(RustTracingLevel.Info, 5, index_arg=9)
            sentry_second_rust_span = sentry_sdk.traces.get_current_span()
            rust_second_rust_span = rust_tracing.spans[5]

            assert rust_second_rust_span == sentry_second_rust_span

            rust_tracing.close_span(5)

            # Ensure the current sentry span was moved back to the parent
            sentry_span_after_close = sentry_sdk.traces.get_current_span()
            assert sentry_span_after_close == sentry_first_rust_span
            assert sentry_span_after_close == rust_first_rust_span

            rust_tracing.close_span(3)

            assert sentry_sdk.traces.get_current_span() == original_sentry_span

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 3

        # Ensure the span metadata is wired up for all spans
        first_span, second_span, _ = spans
        assert first_span["attributes"]["sentry.op"] == "function"
        assert (
            first_span["attributes"]["sentry.origin"]
            == "auto.function.rust_tracing.test_nested_on_new_span_on_close"
        )
        assert first_span["name"] == "_bindings::fibonacci"
        assert second_span["attributes"]["sentry.op"] == "function"
        assert (
            second_span["attributes"]["sentry.origin"]
            == "auto.function.rust_tracing.test_nested_on_new_span_on_close"
        )
        assert second_span["name"] == "_bindings::fibonacci"

        # Ensure the spans were opened/closed appropriately
        assert first_span["start_timestamp"] is not None
        assert first_span["end_timestamp"] is not None
        assert second_span["start_timestamp"] is not None
        assert second_span["end_timestamp"] is not None

        # Ensure the extra data from Rust is hooked up in both spans
        first_span_data = first_span["attributes"]
        assert first_span_data["use_memoized"]
        assert first_span_data["index"] == 9
        assert first_span_data["version"] == "None"

        second_span_data = second_span["attributes"]
        assert second_span_data["use_memoized"]
        assert second_span_data["index"] == 10
        assert second_span_data["version"] == "None"
    else:
        events = capture_events()
        with start_transaction():
            original_sentry_span = sentry_sdk.get_current_span()

            rust_tracing.new_span(RustTracingLevel.Info, 3, index_arg=10)
            sentry_first_rust_span = sentry_sdk.get_current_span()
            rust_first_rust_span = rust_tracing.spans[3]

            # Use a different `index_arg` value for the inner span to help
            # distinguish the two at the end of the test
            rust_tracing.new_span(RustTracingLevel.Info, 5, index_arg=9)
            sentry_second_rust_span = sentry_sdk.get_current_span()
            rust_second_rust_span = rust_tracing.spans[5]

            assert rust_second_rust_span == sentry_second_rust_span

            rust_tracing.close_span(5)

            # Ensure the current sentry span was moved back to the parent
            sentry_span_after_close = sentry_sdk.get_current_span()
            assert sentry_span_after_close == sentry_first_rust_span
            assert sentry_span_after_close == rust_first_rust_span

            rust_tracing.close_span(3)

            assert sentry_sdk.get_current_span() == original_sentry_span

        (event,) = events
        assert len(event["spans"]) == 2

        # Ensure the span metadata is wired up for all spans
        first_span, second_span = event["spans"]
        assert first_span["op"] == "function"
        assert (
            first_span["origin"]
            == "auto.function.rust_tracing.test_nested_on_new_span_on_close"
        )
        assert first_span["description"] == "_bindings::fibonacci"
        assert second_span["op"] == "function"
        assert (
            second_span["origin"]
            == "auto.function.rust_tracing.test_nested_on_new_span_on_close"
        )
        assert second_span["description"] == "_bindings::fibonacci"

        # Ensure the spans were opened/closed appropriately
        assert first_span["start_timestamp"] is not None
        assert first_span["timestamp"] is not None
        assert second_span["start_timestamp"] is not None
        assert second_span["timestamp"] is not None

        # Ensure the extra data from Rust is hooked up in both spans
        first_span_data = first_span["data"]
        assert first_span_data["use_memoized"]
        assert first_span_data["index"] == 10
        assert first_span_data["version"] is None

        second_span_data = second_span["data"]
        assert second_span_data["use_memoized"]
        assert second_span_data["index"] == 9
        assert second_span_data["version"] is None


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_new_span_without_transaction(sentry_init, span_streaming):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_new_span_without_transaction", rust_tracing.set_layer_impl
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        assert sentry_sdk.traces.get_current_span() is None

        # In streaming mode we do not create an orphan root segment when there
        # is no active span
        rust_tracing.new_span(RustTracingLevel.Info, 3)
        assert sentry_sdk.traces.get_current_span() is None
    else:
        assert sentry_sdk.get_current_span() is None

        # Should still create a span hierarchy, it just will not be under a txn
        rust_tracing.new_span(RustTracingLevel.Info, 3)
        current_span = sentry_sdk.get_current_span()
        assert current_span is not None
        assert current_span.containing_transaction is None


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_event_exception(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_event_exception",
        rust_tracing.set_layer_impl,
        event_type_mapping=_test_event_type_mapping,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("event")
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Exception
            rust_tracing.event(RustTracingLevel.Error, 3)

            rust_tracing.close_span(3)

        (exc,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Exception
            rust_tracing.event(RustTracingLevel.Error, 3)

            rust_tracing.close_span(3)

        assert len(events) == 2
        exc, _tx = events

    assert exc["level"] == "error"
    assert exc["logger"] == "_bindings"
    assert exc["message"] == "Getting the 10th fibonacci number"
    assert exc["breadcrumbs"]["values"] == []

    location_context = exc["contexts"]["rust_tracing_location"]
    assert location_context["module_path"] == "_bindings"
    assert location_context["file"] == "src/lib.rs"
    assert location_context["line"] == 23

    field_context = exc["contexts"]["rust_tracing_fields"]
    assert field_context["message"] == "Getting the 10th fibonacci number"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_event_breadcrumb(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_event_breadcrumb",
        rust_tracing.set_layer_impl,
        event_type_mapping=_test_event_type_mapping,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("event")
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Breadcrumb
            rust_tracing.event(RustTracingLevel.Info, 3)

            rust_tracing.close_span(3)
            capture_message("test message")

        (message,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Breadcrumb
            rust_tracing.event(RustTracingLevel.Info, 3)

            rust_tracing.close_span(3)
            capture_message("test message")

        assert len(events) == 2
        message, _tx = events

    breadcrumbs = message["breadcrumbs"]["values"]
    assert len(breadcrumbs) == 1
    assert breadcrumbs[0]["level"] == "info"
    assert breadcrumbs[0]["message"] == "Getting the 10th fibonacci number"
    assert breadcrumbs[0]["type"] == "default"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_event_event(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_event_event",
        rust_tracing.set_layer_impl,
        event_type_mapping=_test_event_type_mapping,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("event")
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Event
            rust_tracing.event(RustTracingLevel.Debug, 3)

            rust_tracing.close_span(3)

        (event,) = (item.payload for item in items if item.type == "event")
    else:
        events = capture_events()
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Mapped to Event
            rust_tracing.event(RustTracingLevel.Debug, 3)

            rust_tracing.close_span(3)

        assert len(events) == 2
        event, _tx = events

    assert event["logger"] == "_bindings"
    assert event["level"] == "debug"
    assert event["message"] == "Getting the 10th fibonacci number"
    assert event["breadcrumbs"]["values"] == []

    location_context = event["contexts"]["rust_tracing_location"]
    assert location_context["module_path"] == "_bindings"
    assert location_context["file"] == "src/lib.rs"
    assert location_context["line"] == 23

    field_context = event["contexts"]["rust_tracing_fields"]
    assert field_context["message"] == "Getting the 10th fibonacci number"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_on_event_ignored(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_on_event_ignored",
        rust_tracing.set_layer_impl,
        event_type_mapping=_test_event_type_mapping,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Ignored
            rust_tracing.event(RustTracingLevel.Trace, 3)

            rust_tracing.close_span(3)

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert spans[1]["is_segment"] is True
    else:
        events = capture_events()
        sentry_sdk.get_isolation_scope().clear_breadcrumbs()

        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            # Ignored
            rust_tracing.event(RustTracingLevel.Trace, 3)

            rust_tracing.close_span(3)

        assert len(events) == 1
        (tx,) = events
        assert tx["type"] == "transaction"
        assert "message" not in tx


@pytest.mark.parametrize("span_streaming", [True, False])
def test_span_filter(
    sentry_init,
    capture_events,
    capture_items,
    span_streaming,
):
    def span_filter(metadata: Dict[str, object]) -> bool:
        return RustTracingLevel(metadata.get("level")) in (
            RustTracingLevel.Error,
            RustTracingLevel.Warn,
            RustTracingLevel.Info,
            RustTracingLevel.Debug,
        )

    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_span_filter",
        initializer=rust_tracing.set_layer_impl,
        span_filter=span_filter,
        include_tracing_fields=True,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )
    if span_streaming:
        items = capture_items("span")
        with sentry_sdk.traces.start_span(name="custom parent"):
            original_sentry_span = sentry_sdk.traces.get_current_span()

            # Span is not ignored
            rust_tracing.new_span(RustTracingLevel.Info, 3, index_arg=10)
            info_span = sentry_sdk.traces.get_current_span()

            # Span is ignored, current span should remain the same
            rust_tracing.new_span(RustTracingLevel.Trace, 5, index_arg=9)
            assert sentry_sdk.traces.get_current_span() == info_span

            # Closing the filtered span should leave the current span alone
            rust_tracing.close_span(5)
            assert sentry_sdk.traces.get_current_span() == info_span

            rust_tracing.close_span(3)
            assert sentry_sdk.traces.get_current_span() == original_sentry_span

        sentry_sdk.flush()
        spans = [item.payload for item in items]
        assert len(spans) == 2
        # The ignored span has index == 9
        assert spans[0]["attributes"]["index"] == 10
    else:
        events = capture_events()
        with start_transaction():
            original_sentry_span = sentry_sdk.get_current_span()

            # Span is not ignored
            rust_tracing.new_span(RustTracingLevel.Info, 3, index_arg=10)
            info_span = sentry_sdk.get_current_span()

            # Span is ignored, current span should remain the same
            rust_tracing.new_span(RustTracingLevel.Trace, 5, index_arg=9)
            assert sentry_sdk.get_current_span() == info_span

            # Closing the filtered span should leave the current span alone
            rust_tracing.close_span(5)
            assert sentry_sdk.get_current_span() == info_span

            rust_tracing.close_span(3)
            assert sentry_sdk.get_current_span() == original_sentry_span

        (event,) = events
        assert len(event["spans"]) == 1
        # The ignored span has index == 9
        assert event["spans"][0]["data"]["index"] == 10


@pytest.mark.parametrize("span_streaming", [True, False])
def test_record(sentry_init, span_streaming):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_record",
        initializer=rust_tracing.set_layer_impl,
        include_tracing_fields=True,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.traces.get_current_span()._to_json()
            assert span_before_record["attributes"]["version"] == "None"

            rust_tracing.record(3)

            span_after_record = sentry_sdk.traces.get_current_span()._to_json()
            assert span_after_record["attributes"]["version"] == "memoized"
    else:
        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.get_current_span().to_json()
            assert span_before_record["data"]["version"] is None

            rust_tracing.record(3)

            span_after_record = sentry_sdk.get_current_span().to_json()
            assert span_after_record["data"]["version"] == "memoized"


@pytest.mark.parametrize("span_streaming", [True, False])
def test_record_in_ignored_span(sentry_init, span_streaming):
    def span_filter(metadata: Dict[str, object]) -> bool:
        # Just ignore Trace
        return RustTracingLevel(metadata.get("level")) != RustTracingLevel.Trace

    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_record_in_ignored_span",
        rust_tracing.set_layer_impl,
        span_filter=span_filter,
        include_tracing_fields=True,
    )
    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.traces.get_current_span()._to_json()
            assert span_before_record["attributes"]["version"] == "None"

            rust_tracing.new_span(RustTracingLevel.Trace, 5)
            rust_tracing.record(5)

            # `on_record()` should not do anything to the current Sentry span if the associated Rust span was ignored
            span_after_record = sentry_sdk.traces.get_current_span()._to_json()
            assert span_after_record["attributes"]["version"] == "None"
    else:
        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.get_current_span().to_json()
            assert span_before_record["data"]["version"] is None

            rust_tracing.new_span(RustTracingLevel.Trace, 5)
            rust_tracing.record(5)

            # `on_record()` should not do anything to the current Sentry span if the associated Rust span was ignored
            span_after_record = sentry_sdk.get_current_span().to_json()
            assert span_after_record["data"]["version"] is None


@pytest.mark.parametrize(
    "send_default_pii, include_tracing_fields, tracing_fields_expected",
    [
        (True, True, True),
        (True, False, False),
        (True, None, True),
        (False, True, True),
        (False, False, False),
        (False, None, False),
    ],
)
@pytest.mark.parametrize("span_streaming", [True, False])
def test_include_tracing_fields(
    sentry_init,
    send_default_pii,
    include_tracing_fields,
    tracing_fields_expected,
    span_streaming,
):
    rust_tracing = FakeRustTracing()
    integration = RustTracingIntegration(
        "test_record",
        initializer=rust_tracing.set_layer_impl,
        include_tracing_fields=include_tracing_fields,
    )

    sentry_init(
        integrations=[integration],
        traces_sample_rate=1.0,
        send_default_pii=send_default_pii,
        trace_lifecycle="stream" if span_streaming else "static",
    )

    if span_streaming:
        with sentry_sdk.traces.start_span(name="custom parent"):
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.traces.get_current_span()._to_json()
            if tracing_fields_expected:
                assert span_before_record["attributes"]["version"] == "None"
            else:
                assert span_before_record["attributes"]["version"] == "[Filtered]"

            rust_tracing.record(3)

            span_after_record = sentry_sdk.traces.get_current_span()._to_json()

            if tracing_fields_expected:
                assert span_after_record["attributes"]["sentry.op"] == "function"
                assert (
                    span_after_record["attributes"]["sentry.origin"]
                    == "auto.function.rust_tracing.test_record"
                )
                assert span_after_record["attributes"]["use_memoized"] is True
                assert span_after_record["attributes"]["version"] == "memoized"
                assert span_after_record["attributes"]["index"] == 10

            else:
                assert span_after_record["attributes"]["sentry.op"] == "function"
                assert (
                    span_after_record["attributes"]["sentry.origin"]
                    == "auto.function.rust_tracing.test_record"
                )
                assert span_after_record["attributes"]["use_memoized"] == "[Filtered]"
                assert span_after_record["attributes"]["version"] == "[Filtered]"
                assert span_after_record["attributes"]["index"] == "[Filtered]"
    else:
        with start_transaction():
            rust_tracing.new_span(RustTracingLevel.Info, 3)

            span_before_record = sentry_sdk.get_current_span().to_json()
            if tracing_fields_expected:
                assert span_before_record["data"]["version"] is None
            else:
                assert span_before_record["data"]["version"] == "[Filtered]"

            rust_tracing.record(3)

            span_after_record = sentry_sdk.get_current_span().to_json()

            if tracing_fields_expected:
                assert span_after_record["data"]["use_memoized"] is True
                assert span_after_record["data"]["version"] == "memoized"
                assert span_after_record["data"]["index"] == 10

            else:
                assert span_after_record["data"]["use_memoized"] == "[Filtered]"
                assert span_after_record["data"]["version"] == "[Filtered]"
                assert span_after_record["data"]["index"] == "[Filtered]"



from typing import TYPE_CHECKING

from sentry_sdk._batcher import Batcher
from sentry_sdk.envelope import Item, PayloadRef
from sentry_sdk.utils import serialize_attribute

if TYPE_CHECKING:
    from typing import Any

    from sentry_sdk._types import Log


class LogBatcher(Batcher["Log"]):
    MAX_BEFORE_FLUSH = 100
    MAX_BEFORE_DROP = 1_000
    FLUSH_WAIT_TIME = 5.0

    TYPE = "log"
    CONTENT_TYPE = "application/vnd.sentry.items.log+json"

    @staticmethod
    def _to_transport_format(item: "Log") -> "Any":
        if "sentry.severity_number" not in item["attributes"]:
            item["attributes"]["sentry.severity_number"] = item["severity_number"]
        if "sentry.severity_text" not in item["attributes"]:
            item["attributes"]["sentry.severity_text"] = item["severity_text"]

        res = {
            "timestamp": int(item["time_unix_nano"]) / 1.0e9,
            "level": str(item["severity_text"]),
            "body": str(item["body"]),
            "attributes": {
                k: serialize_attribute(v) for (k, v) in item["attributes"].items()
            },
        }

        if item.get("trace_id") is not None:
            res["trace_id"] = item["trace_id"]

        if item.get("span_id") is not None:
            res["span_id"] = item["span_id"]

        return res

    def _record_lost(self, item: "Log") -> None:
        # Construct log envelope item without sending it to report lost bytes
        log_item = Item(
            type=self.TYPE,
            content_type=self.CONTENT_TYPE,
            headers={
                "item_count": 1,
            },
            payload=PayloadRef(
                json={
                    "version": 2,
                    "items": [self._to_transport_format(item)],
                }
            ),
        )

        self._record_lost_func(
            reason="queue_overflow",
            data_category="log_item",
            item=log_item,
            quantity=1,
        )
